import os

from scm import plams
from typing import Dict, Optional, TYPE_CHECKING
from tcutility import data, formula, log, molecule, results, spell_check
from tcutility.errors import TCCompDetailsError, TCJobError
from tcutility.job.ams import AMSJob
from tcutility.job.generic import Job

j = os.path.join

if TYPE_CHECKING:
    import pyfmo

class ADFJob(AMSJob):
    def __init__(self, *args, **kwargs):
        self.settings = results.Result()
        self._functional = None
        self._core = None
        self.solvent("vacuum")
        self.basis_set("TZ2P")
        self.quality("Good")
        self.SCF(thresh=1e-8)
        self.single_point()

        super().__init__(*args, **kwargs)

    def __str__(self):
        return f"{self._task}({self._functional}/{self._basis_set}), running in {os.path.join(os.path.abspath(self.rundir), self.name)}"

    def basis_set(self, typ: str = "TZ2P", core: str = "None"):
        """
        Set the basis-set type and frozen core approximation for this calculation.

        Args:
            typ: the type of basis-set to use. Default is TZ2P.
            core: the size of the frozen core approximation. Default is None.

        Raises:
            ValueError: if the basis-set name or core is incorrect.

        .. note:: If the selected functional is the r2SCAN-3c functional, then the basis-set will be set to mTZ2P.

        .. seealso::
            :mod:`tcutility.data.basis_sets` for an overview of the available basis-sets in ADF.
        """
        spell_check.check(typ, data.basis_sets.available_basis_sets["ADF"], ignore_case=True)
        spell_check.check(core, ["None", "Small", "Large"], ignore_case=True)
        if self._functional == "r2SCAN-3c" and typ != "mTZ2P":
            log.warn(f"Basis set {typ} is not allowed with r2SCAN-3c, switching to mTZ2P.")
            typ = "mTZ2P"
        self._basis_set = typ
        self._core = core
        self.settings.input.adf.basis.type = typ
        self.settings.input.adf.basis.core = core

    def spin_polarization(self, val: int):
        """
        Set the spin-polarization of the system. If the value is not zero the calculation will also be unrestricted.
        """
        self.settings.input.adf.SpinPolarization = val
        if val != 0:
            self.settings.input.adf.Unrestricted = "Yes"

    def multiplicity(self, val: int):
        """
        Set the multiplicity of the system. If the value is not one the calculation will also be unrestricted.
        We use the following values:

        1) singlet
        2) doublet
        3) triplet
        4) ...

        The multiplicity is equal to 2*S+1 for spin-polarization S.
        """
        self.settings.input.adf.SpinPolarization = (val - 1) // 2
        if val != 1:
            self.settings.input.adf.Unrestricted = "Yes"

    def unrestricted(self, val: bool):
        """
        Whether the calculation should be unrestricted.
        """
        self.settings.input.adf.Unrestricted = "Yes" if val else "No"

    def occupations(self, strategy: str):
        """
        Set the orbital filling strategy for ADF.

        Args:
            strategy: the name of the filling strategy. This can contain multiple of the options allowed.

        .. seealso::
            The SCM documentation can be found at https://www.scm.com/doc/ADF/Input/Electronic_Configuration.html#aufbau-smearing-freezing
        """
        self.settings.input.adf.Occupations = strategy

    def quality(self, val: str = "Good"):
        """
        Set the numerical quality of the calculation.

        Args:
            val: the numerical quality value to set to. This is the same as the ones used in the ADF GUI. Defaults to Good.

        Raises:
            ValueError: if the quality value is incorrect.
        """
        spell_check.check(val, ["Basic", "Normal", "Good", "VeryGood", "Excellent"], ignore_case=True)
        self.settings.input.adf.NumericalQuality = val

    def SCF_convergence(self, thresh: float = 1e-8):
        """
        Set the SCF convergence criteria for the job.

        Args:
            thresh: the convergence criteria for the SCF procedure. Defaults to 1e-8.

        .. deprecated:: 0.9.2
            Please use :meth:`ADFJob.SCF` instead of this method.
        """
        log.warn("This method has been deprecated, please use ADFJob.SCF instead.")
        self.SCF(thresh=thresh)

    def SCF(self, iterations: int = 300, thresh: float = 1e-8):
        """
        Set the SCF settings for this calculations.

        Args:
            iterations: number of iterations to perform for this calculation. Defaults to 300.
            thresh: the convergence criteria for the SCF procedure. Defaults to 1e-8.

        .. note::
            When setting the number of iterations to 0 or 1 the ``AlwaysClaimSuccess`` key will also be set to ``Yes``.
            This is to prevent the job from being flagged as a failure when reading it using :mod:`tcutility.results`.
        """
        self.settings.input.adf.SCF.Iterations = iterations
        self.settings.input.adf.SCF.Converge = thresh

        if iterations in [0, 1]:
            self.settings.input.ams.EngineDebugging.AlwaysClaimSuccess = "Yes"

    def functional(self, funtional_name: str, dispersion: str = None):
        """
        Set the functional to be used by the calculation. This also sets the dispersion if it is specified in the functional name.

        Args:
            funtional_name: the name of the functional. The value can be the same as the ones used in the ADF GUI.
            dispersion: dispersion setting to use with the functional. This is used when you want to use a functional from LibXC.

        Raises:
            ValueError: if the functional name is not recognized.

        .. note:: Setting the functional to r2SCAN-3c will automatically set the basis-set to mTZ2P.

        .. seealso::
            :mod:`tcutility.data.functionals` for an overview of the available functionals in ADF.
        """
        # before adding the new functional we should clear any previous functional settings
        self.settings.input.adf.pop("XC", None)

        functional = funtional_name.strip()
        functional = functional.replace("-D4(EEQ)", "-D4")  # D4(EEQ) and D4 are the same, unlike D3 and D3(BJ)
        self._functional = functional

        if functional == "r2SCAN-3c" and self._basis_set != "mTZ2P":
            log.info(f"Switching basis set from {self._basis_set} to mTZ2P for r2SCAN-3c.")
            self.basis_set("mTZ2P")

        if functional == "SSB-D":
            raise TCCompDetailsError(section="Functional", message='There are two functionals called SSB-D, please use "GGA:SSB-D" or "MetaGGA:SSB-D".')

        if not data.functionals.get(functional):
            raise TCCompDetailsError(section="Functional", message=f"XC-functional {functional} not found.")
        else:
            func = data.functionals.get(functional)
            self.settings.input.adf.update(func.adf_settings)

    def relativity(self, level: str = "Scalar"):
        """
        Set the treatment of relativistic effects for this calculation.

        Args:
            level: the level to set. Can be the same as the values in the ADF GUI and documentation. By default it is set to Scalar.

        Raises:
            ValueError: if the relativistic correction level is not correct.
        """
        spell_check.check(level, ["Scalar", "None", "Spin-Orbit"], ignore_case=True)
        self.settings.input.adf.relativity.level = level

    def solvent(self, name: str = None, eps: float = None, rad: float = None, use_klamt: bool = False, radii: Optional[Dict[str, float]] = None):
        """
        Model solvation using COSMO for this calculation.

        Args:
            name: the name of the solvent you want to use. Please see the ADF manual for a list of available solvents.
            eps: the dielectric constant of your solvent. You can use this in place of the solvent name if you need more control over COSMO.
            rad: the radius of the solvent molecules. You can use this in place of the solvent name if you need more control over COSMO.
            use_klamt: whether to use the klamt atomic radii. This is usually used when you have charged species (?).
            radii: the atomic radii to use for the COSMO calculation. A dictionary such as {Cl: 1.04} will be converted to Cl = 0.01 in the input script. This gets (partially) overwritten if use_klamt is enables.

        Raises:
            ValueError: if the solvent name is given, but incorrect.

        .. seealso::
            :mod:`tcutility.data.cosmo` for an overview of the available solvent names and formulas.
        """
        if name:
            spell_check.check(name, data.cosmo.available_solvents, ignore_case=True, insertion_cost=0.3)
            self._solvent = name
        else:
            self._solvent = f"COSMO(eps={eps} rad={rad})"

        if name == "vacuum":
            self.settings.input.adf.pop("Solvation", None)
            return

        self.settings.input.adf.Solvation.Surf = "Delley"
        solv_string = ""
        if name:
            solv_string += f"name={name} "
        else:
            self.settings.input.adf.Solvation.Solv = f"eps={eps} rad={rad} "
        if use_klamt:
            solv_string += "cav0=0.0 cav1=0.0"
        else:
            solv_string += "cav0=0.0 cav1=0.0067639"
        self.settings.input.adf.Solvation.Solv = solv_string

        self.settings.input.adf.Solvation.Charged = "method=CONJ corr"
        self.settings.input.adf.Solvation["C-Mat"] = "POT"
        self.settings.input.adf.Solvation.SCF = "VAR ALL"
        self.settings.input.adf.Solvation.CSMRSP = None

        if radii is not None:
            self.settings.input.adf.solvation.radii = radii

        if use_klamt:
            radii = {"H": 1.30, "C": 2.00, "N": 1.83, "O": 1.72, "F": 1.72, "Si": 2.48, "P": 2.13, "S": 2.16, "Cl": 2.05, "Br": 2.16, "I": 2.32}
            self.settings.input.adf.solvation.radii = radii

    def symmetry(self, group: str):
        """
        Specify the symmetry group to be used by ADF.

        Args:
            group: the symmetry group to be used.
        """
        self.settings.input.adf.Symmetry = group

    def _check_job(self):
        # if we have spinpolarization we do not want to use DFTB to calculate the initial hessian
        if self.settings.input.adf.SpinPolarization != 0 and self.settings.input.ams.GeometryOptimization.InitialHessian.Type == "CalculateWithFastEngine":
            # we simply remove it from the geometryoptimization block
            self.settings.input.ams.GeometryOptimization.pop("InitialHessian", None)


class ADFFragmentJob(ADFJob):
    def __init__(self, *args, **kwargs):

        self.decompose_elstat = kwargs.pop('decompose_elstat', False)
        self.counter_poise = kwargs.pop('counter_poise', False)
        self.scf0_calculation = kwargs.pop('sfo0_calculation', False)
        self.child_jobs = {}
        # self.child_jobs_no_electrons = {}
        super().__init__(*args, **kwargs)
        self.name = "EDA"

        # by default print the fock matrix
        # self.settings.input.adf.print = "FmatSFO"

    def add_fragment(self, mol: plams.Molecule, name: str = None, charge: int = 0, spin_polarization: int = 0):
        """
        Add a fragment to this job. Optionally give the name, charge and spin-polarization of the fragment as well.

        Args:
            mol: the molecule corresponding to the fragment.
            name: the name of the fragment. By default it will be set to ``fragment{N+1}`` if ``N`` is the number of fragments already present.
            charge: the charge of the fragment to be added.
            spin_polarization: the spin-polarization of the fragment to be added.
        """
        # in case of giving fragments as indices we dont want to add the fragment to the molecule later
        add_frag_to_mol = True
        # we can be given a list of atoms
        if isinstance(mol, list) and isinstance(mol[0], plams.Atom):
            mol_ = plams.Molecule()
            [mol_.add_atom(atom) for atom in mol]
            mol = mol_

        # or a list of integers
        if isinstance(mol, list) and isinstance(mol[0], int):
            if not self._molecule:
                log.error(f"Trying to add fragment based on atom indices, but main job does not have a molecule yet. Call the {self.__class__.__name__}.molecule method to add one.")
                return
            mol_ = plams.Molecule()
            [mol_.add_atom(self._molecule[i]) for i in mol]
            mol = mol_.copy()
            add_frag_to_mol = False

        # check if the atoms in the new fragment are already present in the other fragments.
        # if it is we should raise an error
        for child in self.child_jobs.values():
            if any((atom.symbol, atom.coords) == (myatom.symbol, myatom.coords) for atom in child._molecule for myatom in mol):
                raise TCJobError(job_class=self.__class__.__name__, message="The atoms in the new fragment are already present in the other fragments.")

        name = name or f"fragment{len(self.child_jobs) + 1}"
        self.child_jobs[name] = ADFJob(test_mode=self.test_mode)
        self.child_jobs[name].molecule(mol)
        if charge:
            self.child_jobs[name].charge(charge)
        if spin_polarization:
            self.child_jobs[name].spin_polarization(spin_polarization)
        setattr(self, name, self.child_jobs[name])

        if not add_frag_to_mol:
            return

        if self._molecule is None:
            self._molecule = self.child_jobs[name]._molecule.copy()
        else:
            for atom in self.child_jobs[name]._molecule.copy():
                if any((atom.symbol, atom.coords) == (myatom.symbol, myatom.coords) for myatom in self._molecule):
                    continue
                self._molecule.add_atom(atom)

    def guess_fragments(self):
        """
        Guess what the fragments are based on data stored in the molecule provided for this job.
        This will automatically set the correct fragment molecules, names, charges and spin-polarizations.

        .. seealso::
            | :func:`tcutility.molecule.guess_fragments` for an explanation of the xyz-file format required to guess the fragments.
            | :meth:`ADFFragmentJob.add_fragment` to manually add a fragment.

        .. note::
            This function will be automatically called if there were no fragments given to this calculation.
        """
        frags = molecule.guess_fragments(self._molecule)
        if frags is None:
            log.error("Could not load fragment data for the molecule.")
            return False

        for fragment_name, fragment in frags.items():
            charge = fragment.flags.get("charge", 0)
            spin_polarization = fragment.flags.get("spinpol", 0)
            self.add_fragment(fragment, fragment_name, charge=charge, spin_polarization=spin_polarization)

        return True

    def remove_virtuals(self, frag=None, subspecies=None, nremove=None):
        """
        Remove virtual orbitals from the fragments.

        Args:
            frag: the fragment to remove virtuals from. If set to ``None`` we remove all virtual orbitals of all fragments.
            subspecies: the symmetry subspecies to remove virtuals from. If set to ``None`` we assume we have ``A`` subspecies.
            nremove: the number of virtuals to remove. If set to ``None`` we will guess the number of virtuals based on the basis-set chosen.
        """
        if frag is None:
            self.settings.input.adf.RemoveAllFragVirtuals = "Yes"
            return

        # if nremove is not given we will get it from the atoms in the fragment
        if nremove is None:
            # guess the virtual numbers only works for non-frozen-core calculations
            if self._core.lower() != "none":
                raise TCJobError(job_class=self.__class__.__name__, message="Cannot guess number of virtual orbitals for calculations with a frozen core.")
            # the basis-set has to be present in the prepared data
            if self._basis_set.lower() not in [bs.lower() for bs in data.basis_sets._number_of_orbitals.keys()]:
                raise TCJobError(job_class=self.__class__.__name__, message=f"Cannot guess number of virtual orbitals for calculations with the {self._basis_set} basis-set.")

            # sum up the number of virtuals per atom in the fragment
            nremove = 0
            for atom in self.child_jobs[frag]._molecule:
                nremove += data.basis_sets.number_of_virtuals(atom.symbol, self._basis_set)
            # positive charge adds a virtual and negative removes a virtual
            nremove += self.child_jobs[frag].settings.input.ams.System.charge or 0

        self.settings.input.adf.setdefault("RemoveFragOrbitals", "")
        self.settings.input.adf.RemoveFragOrbitals += f"""
    {frag}
      {subspecies or 'A'} {nremove}
    SubEnd
  End
        """

    def frag_occupations(self, frag=None, subspecies=None, alpha=None, beta=None):
        """
        Set the occupations of the fragments.

        Args:
            frag: the fragment to set the occupations for.
            subspecies: the symmetry subspecies to set the occupations for. If set to ``None`` we assume we have ``A`` subspecies.
            alpha: the number of alpha electrons. If set to ``None`` we will guess the number of electrons based on the spin-polarization set.
            beta: the number of beta electrons. If set to ``None`` we will guess the number of electrons based on the spin-polarization set.
        """

        child_job = self.child_jobs[frag]

        if alpha is None and beta is None:
            spinpol = child_job.settings.input.adf.SpinPolarization or 0
            charge = child_job.settings.input.ams.system.charge or 0
            print(child_job._molecule)
            nelectrons = sum(atom.atnum for atom in child_job._molecule) - charge
            alpha = nelectrons // 2 + spinpol
            beta  = nelectrons // 2

        self.settings.input.adf.setdefault("FragOccupations", "")
        self.settings.input.adf.FragOccupations = self.settings.input.adf.FragOccupations.replace(" End", "")
        self.settings.input.adf.FragOccupations += f"""
    {frag}
      {subspecies or 'A'} {alpha} // {beta}
    SubEnd
  End
        """

    def run(self):
        """
        Run the ``ADFFragmentJob``. This involves setting up the calculations for each fragment as well as the parent job.
        It will also submit a calculation with the SCF iterations set to 0 in order to facilitate investigation of the field effects using PyOrb.
        """
        # check if the user defined fragments for this job

        if len(self.child_jobs) == 0:
            log.warn("Fragments were not specified yet, trying to read them from the xyz file ...")

            # if they did not define the fragments, try to guess them using the xyz-file
            if not self.guess_fragments():
                log.error("Please specify the fragments using ADFFragmentJob.add_fragment or specify them in the xyz file.")
                raise

        mol_str = " + ".join([formula.molecule(child._molecule) for child in self.child_jobs.values()])
        log.flow(f"ADFFragmentJob [{mol_str}]", ["start"])
        # obtain some system wide properties of the molecules
        charge = sum([child.settings.input.ams.System.charge or 0 for child in self.child_jobs.values()])
        unrestricted = any([(child.settings.input.adf.Unrestricted or "no").lower() == "yes" for child in self.child_jobs.values()])
        spinpol = sum([child.settings.input.adf.SpinPolarization or 0 for child in self.child_jobs.values()])
        log.flow(f"Level:             {self._functional}/{self._basis_set}")
        log.flow(f"Solvent:           {self._solvent}")
        log.flow(f"Charge:            {charge}", ["straight"])
        log.flow(f"Unrestricted:      {unrestricted}", ["straight"])
        log.flow(f"Spin-Polarization: {spinpol}", ["straight"])
        log.flow()

        # this job and all its children should have the same value for unrestricted
        # [child.unrestricted(unrestricted) for child in self.child_jobs.values()]

        # propagate the post- and preambles to the child_jobs
        [child.add_preamble(preamble) for preamble in self._preambles for child in self.child_jobs.values()]
        [child.add_postamble(postamble) for postamble in self._postambles for child in self.child_jobs.values()]

        # we now update the child settings with the parent settings
        # this is because we have to propagate settings such as functionals, basis sets etc.
        sett = self.settings.as_plams_settings()  # first create a plams settings object
        # same for the children
        child_setts = {name: child.settings.as_plams_settings() for name, child in self.child_jobs.items()}
        # update the children using the parent settings
        [child_sett.update(sett) for child_sett in child_setts.values()]
        [child_sett.input.adf.pop("RemoveFragOrbitals", None) for child_sett in child_setts.values()]
        [child_sett.input.adf.pop("RemoveAllFragVirtuals", None) for child_sett in child_setts.values()]
        [child_sett.input.adf.pop("FragOccupations", None) for child_sett in child_setts.values()]
        # same for sbatch settings
        [child.sbatch(**self._sbatch) for child in self.child_jobs.values()]

        # now set the charge, spinpol, unrestricted for the parent
        if charge:
            self.charge(charge)
        if spinpol:
            self.spin_polarization(spinpol)
        if unrestricted:
            self.unrestricted(unrestricted)
            self.settings.input.adf.UnrestrictedFragments = "Yes"

        elstat_jobs = {}

        # now we are going to run each child job
        for i, (child_name, child) in enumerate(self.child_jobs.items(), start=1):
            # the child name will be prepended with SP showing that it is the singlepoint calculation
            child.name = f"frag_{child_name}"
            child.rundir = j(self.rundir, self.name)

            # # add the path to the child adf.rkf file as a dependency to the parent job
            self.settings.input.adf.fragments[child_name] = j(child.workdir, "adf.rkf")

            # recast the plams.Settings object into a Result object as that is what run expects
            child.settings = results.Result(child_setts[child_name])

            log.flow(f'Fragment ({i}/{len(self.child_jobs)}) {child_name} [{formula.molecule(child._molecule)}]', ['split'])
            log.flow(f'Charge:            {child.settings.input.ams.System.charge or 0}', ['straight', 'straight'])
            log.flow(f'Spin-Polarization: {child.settings.input.adf.SpinPolarization or 0}', ['straight', 'straight'])
            log.flow(f'Work dir:          {child.workdir}', ['straight', 'straight'])

            if child.can_skip():
                log.flow(log.Emojis.warning + " Already ran, skipping", ["straight", "end"])
                log.flow()
            else:
                log.flow(log.Emojis.good + " Submitting", ["straight", "end"])
                [child._sbatch.pop(key, None) for key in ["D", "chdir", "J", "job_name", "o", "output"]]
                child.run()
                self.dependency(child)
                log.flow(f"SlurmID:  {child.slurm_job_id}", ["straight", "skip", "end"])
                log.flow()

            if self.decompose_elstat:
                child_STOFIT = ADFJob(child)
                child_STOFIT.name += "_STOFIT"
                elstat_jobs[child_STOFIT.name] = child_STOFIT
                child_STOFIT.settings.input.adf.STOFIT = ""
                child_STOFIT.settings.input.adf.PRINT += " Elstat"
                child_STOFIT.settings.input.adf.pop("NumericalQuality")
                child_STOFIT.settings.input.adf.BeckeGrid.Quality = "Excellent"

                log.flow(f'Fragment ({i}/{len(self.child_jobs)}) {child_name} [{formula.molecule(child._molecule)}] with STOFIT', ['split'])
                log.flow(f'Charge:            {child_STOFIT.settings.input.ams.System.charge or 0}', ['straight', 'straight'])
                log.flow(f'Spin-Polarization: {child_STOFIT.settings.input.adf.SpinPolarization or 0}', ['straight', 'straight'])
                log.flow(f'Work dir:          {child_STOFIT.workdir}', ['straight', 'straight'])

                if child_STOFIT.can_skip():
                    log.flow(log.Emojis.warning + " Already ran, skipping", ["straight", "end"])
                    log.flow()
                else:
                    log.flow(log.Emojis.good + " Submitting", ["straight", "end"])
                    [child_STOFIT._sbatch.pop(key, None) for key in ["D", "chdir", "J", "job_name", "o", "output"]]
                    child_STOFIT.run()
                    self.dependency(child_STOFIT)
                    log.flow(f"SlurmID:  {child_STOFIT.slurm_job_id}", ["straight", "skip", "end"])
                    log.flow()

                child_NoElectrons = ADFJob(child)
                child_NoElectrons.name += "_NoElectrons"
                elstat_jobs[child_NoElectrons.name] = child_NoElectrons

                child_NoElectrons.settings.input.adf.STOFIT = ''
                child_NoElectrons.settings.input.adf.PRINT += ' Elstat'
                child_NoElectrons.charge(molecule.number_of_electrons(child_NoElectrons._molecule))
                child_NoElectrons.spin_polarization(0)
                child_NoElectrons.settings.input.adf.pop("NumericalQuality")
                child_NoElectrons.settings.input.adf.BeckeGrid.Quality = "Excellent"


                log.flow(f'Fragment ({i}/{len(self.child_jobs)}) {child_name} [{formula.molecule(child._molecule)}] without Electrons', ['split'])
                log.flow(f'Charge:            {child_NoElectrons.settings.input.ams.System.charge or 0}', ['straight', 'straight'])
                log.flow(f'Spin-Polarization: {child_NoElectrons.settings.input.adf.SpinPolarization or 0}', ['straight', 'straight'])
                log.flow(f'Work dir:          {child_NoElectrons.workdir}', ['straight', 'straight'])

                if child_NoElectrons.can_skip():
                    log.flow(log.Emojis.warning + " Already ran, skipping", ["straight", "end"])
                    log.flow()
                else:
                    log.flow(log.Emojis.good + " Submitting", ["straight", "end"])
                    [child_NoElectrons._sbatch.pop(key, None) for key in ["D", "chdir", "J", "job_name", "o", "output"]]
                    child_NoElectrons.run()
                    self.dependency(child_NoElectrons)
                    log.flow(f"SlurmID:  {child_NoElectrons.slurm_job_id}", ["straight", "skip", "end"])
                    log.flow()

        # in the parent job the atoms should have the region and adf.f defined as options
        atom_lines = []
        # for each atom we check which child it came from
        for atom in self._molecule:
            for child_name, child in self.child_jobs.items():
                for childatom in child._molecule:
                    # we check by looking at the symbol and coordinates of the atom
                    if (atom.symbol, atom.x, atom.y, atom.z) == (childatom.symbol, childatom.x, childatom.y, childatom.z):
                        # now write the symbol and coords as a string with the correct suffix
                        atom_lines.append(f"\t\t{atom.symbol} {atom.x} {atom.y} {atom.z} region={child_name} adf.f={child_name}")

        old_name = self.name
        # write the atoms block as a string with new line characters
        self.settings.input.ams.system.atoms = ("\n" + "\n".join(atom_lines) + "\n\tEnd").expandtabs(4)
        # set the _molecule to None, otherwise it will overwrite the atoms block
        self._molecule = None
        # run this job
        self.rundir = j(self.rundir, old_name)
        self.name = "complex"
        log.flow(log.Emojis.good + " Submitting parent job", ["split"])

        super().run()
        log.flow(f"SlurmID: {self.slurm_job_id}", ["straight", "end"])
        log.flow()

        # also do the calculation with SCF cycles set to 1 if desired
        if self.scf0_calculation:
            old_iters = self.settings.input.adf.SCF.Iterations or 300
            self.SCF(iterations=0)
            # we must repopulate the sbatch settings for the new run
            [self._sbatch.pop(key, None) for key in ["D", "chdir", "J", "job_name", "o", "output"]]
            self.name = "complex_SCF0"
            log.flow(log.Emojis.good + " Submitting extra job with 0 SCF iterations", ["split"])

            super().run()
            log.flow(f"SlurmID: {self.slurm_job_id}", ["straight", "end"])
            log.flow()

            # reset the SCF iterations
            self.SCF(iterations=old_iters)

        # also do the calculation with no electrons on the fragments if the user requested a elstat decomposition
        if self.decompose_elstat:
            frag_names = self.child_jobs.keys()
            self.settings.input.adf.pop("NumericalQuality")
            self.settings.input.adf.BeckeGrid.Quality = "Excellent"
            [self._sbatch.pop(key, None) for key in ["D", "chdir", "J", "job_name", "o", "output"]]
            self.name = "complex_STOFIT"

            # set the region and fragment files correctly
            atom_lines_ = []
            for line in atom_lines:
                for frag in frag_names:
                    line = line.replace(frag, f"{frag}_STOFIT")
                atom_lines_.append(line)

            self.settings.input.ams.system.atoms = ("\n" + "\n".join(atom_lines_) + "\n\tEnd").expandtabs(4)
            # set the fragment references correctly
            self.settings.input.adf.pop("fragments")
            for frag_name in frag_names:
                self.settings.input.adf.fragments[frag_name + "_STOFIT"] = j(elstat_jobs["frag_" + frag_name + "_STOFIT"].workdir, "adf.rkf")
            self.settings.input.adf.STOFIT = ""
            self.settings.input.adf.PRINT += " Elstat"
            [self._sbatch.pop(key, None) for key in ["D", "chdir", "J", "job_name", "o", "output"]]
            log.flow(log.Emojis.good + " Submitting complex with STOFIT", ["split"])
            super().run()
            log.flow(f"SlurmID: {self.slurm_job_id}", ["straight", "end"])
            log.flow()

            for frag in frag_names:
                # other_frags stores fragment names for fragments that keep their electrons
                other_frags = [frag_ for frag_ in frag_names if frag_ != frag]

                # we must repopulate the sbatch settings for the new run
                [self._sbatch.pop(key, None) for key in ["D", "chdir", "J", "job_name", "o", "output"]]
                self.name = f"complex_{frag}_NoElectrons"

                # set the region and fragment files correctly
                atom_lines_ = []
                for line in atom_lines:
                    line = line.replace(frag, f"{frag}_NoElectrons")
                    for other_frag in other_frags:
                        line = line.replace(other_frag, f"{other_frag}_STOFIT")
                    atom_lines_.append(line)
                self.settings.input.ams.system.atoms = ("\n" + "\n".join(atom_lines_) + "\n\tEnd").expandtabs(4)

                # set the fragment references correctly
                self.settings.input.adf.pop("fragments")
                self.settings.input.adf.fragments[frag + "_NoElectrons"] = j(elstat_jobs["frag_" + frag + "_NoElectrons"].workdir, "adf.rkf")
                for other_frag in other_frags:
                    self.settings.input.adf.fragments[other_frag + "_STOFIT"] = j(elstat_jobs["frag_" + other_frag + "_STOFIT"].workdir, "adf.rkf")

                other_charge = sum([elstat_jobs["frag_" + other_frag + "_STOFIT"].settings.input.ams.System.charge for other_frag in other_frags])
                total_charge = other_charge + (elstat_jobs["frag_" + frag + "_NoElectrons"].settings.input.ams.System.charge)
                self.charge(total_charge)

                other_spin_polarization = sum([elstat_jobs["frag_" + other_frag + "_STOFIT"].settings.input.adf.SpinPolarization for other_frag in other_frags])
                total_spin_polarization = other_spin_polarization + (elstat_jobs["frag_" + frag + "_NoElectrons"].settings.input.adf.SpinPolarization)
                self.spin_polarization(total_spin_polarization)

                log.flow(log.Emojis.good + f" Submitting complex with 0 electrons in fragment {frag}", ["split"])
                [self._sbatch.pop(key, None) for key in ["D", "chdir", "J", "job_name", "o", "output"]]
                super().run()
                log.flow(f"SlurmID: {self.slurm_job_id}", ["straight", "end"])
                log.flow()

        if self.counter_poise:
            self.settings.input.ams.EngineDebugging.pop('AlwaysClaimSuccess', None)
            self.settings.input.adf.pop('fragments', None)
            for frag, frag_job in self.child_jobs.items():
                atoms = [atom for job in self.child_jobs.values() for atom in job._molecule]

                # in the parent job the atoms should have the region and adf.f defined as options
                atom_lines = []
                # for each atom we check which child it came from
                for atom in atoms:
                    for child_name, child in self.child_jobs.items():
                        for childatom in child._molecule:
                            # we check by looking at the symbol and coordinates of the atom
                            if (atom.symbol, atom.x, atom.y, atom.z) == (childatom.symbol, childatom.x, childatom.y, childatom.z):
                                # now write the symbol and coords as a string with the correct suffix and ghost indicator
                                if child_name == frag:
                                    atom_lines.append(f'\t\t{atom.symbol} {atom.x} {atom.y} {atom.z}')
                                else:
                                    atom_lines.append(f'\t\tGh.{atom.symbol} {atom.x} {atom.y} {atom.z}')

                # write the atoms block as a string with new line characters
                self.settings.input.ams.system.atoms = ("\n" + "\n".join(atom_lines) + "\n\tEnd").expandtabs(4)
                self.spin_polarization(frag_job.settings.input.adf.SpinPolarization or 0)
                self.charge(frag_job.settings.input.ams.System.charge or 0)
                self.unrestricted((frag_job.settings.input.adf.Unrestricted or 'no').lower() == 'yes')

                # we must repopulate the sbatch settings for the new run
                self.name = f'complex_{frag}_Ghost'
                log.flow(log.Emojis.good + f' Submitting {frag} with the basis-set of the complex', ['split'])
                [self._sbatch.pop(key, None) for key in ["D", "chdir", "J", "job_name", "o", "output"]]
                super().run()
                log.flow(f'SlurmID: {self.slurm_job_id}', ['straight', 'end'])
                log.flow()

        log.flow(log.Emojis.finish + ' Done, bye!', ['startinv'])



class DensfJob(Job):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.settings = results.Result()
        self.rundir = "tmp"
        self.name = "densf"
        self.gridsize()
        self._mos = []
        self._sfos = []
        self._extras = []
        self.settings.ADFFile = None

    def __str__(self):
        return f"Densf({self.target}), running in {self.workdir}"

    def gridsize(self, size="medium"):
        """
        Set the size of the grid to be used by Densf.

        Args:
            size: either "coarse", "medium", "fine". Defaults to "medium".
        """
        spell_check.check(size, ["coarse", "medium", "fine"], ignore_case=True)
        self.settings.grid = size

    def grid(self, args):
        self.settings.grid = '\n' + '\n'.join(args)

    def orbital(self, orbital: "pyfmo.orbitals.sfo.SFO" or "pyfmo.orbitals.mo.MO"):  # noqa: F821
        """
        Add a PyOrb orbital for Densf to calculate.
        """
        import pyfmo

        if isinstance(orbital, (pyfmo.orbitals.sfo.SFO, pyfmo.orbitals2.objects.SFO)):
            self._sfos.append(orbital)
        elif isinstance(orbital, (pyfmo.orbitals.mo.MO, pyfmo.orbitals2.objects.MO)):
            self._mos.append(orbital)
        else:
            raise TCJobError(job_class=self.__class__.__name__, message=f"Unknown object {orbital} of type{type(orbital)}. It should be a pyfmo.orbitals.sfo.SFO or pyfmo.orbitals.mos.MO object.")

        # check if the ADFFile is the same for all added orbitals
        if self.settings.ADFFile is None:
            self.settings.ADFFile = orbital.kfpath

        elif self.settings.ADFFile != orbital.kfpath:
            raise TCJobError(job_class=self.__class__.__name__, message="RKF file that was previously set not the same as the one being set now. Please start a new job for each RKF file.")


    def density(self, orbitals: 'pyfmo.orbitals.Orbitals'):  # noqa: F821

        # check if the ADFFile is the same for all added orbitals
        if self.settings.ADFFile is None:
            self.settings.ADFFile = orbitals.kfpath

        elif self.settings.ADFFile != orbitals.kfpath:
            raise ValueError("RKF file that was previously set not the same as the one being set now. Please start a new job for each RKF file.")

        self._extras.append("Density SCF")

    def NCI(self, density: str = 'both', rhovdw=0.02, rdg=0.5):
        '''
        Setup calculation of NCI values.

        Args:
            density: the density to calculate for, either "FIT" or  "BOTH", default is "BOTH".
            rhovdw: threshold of density for detection of weak interaction regions, default is `0.02`.
            rdg: threshold of reduced density gradient, default is `0.5`.
        '''
        self._extras.append(f'NCI {density} RHOVDW={rhovdw} RDG={rdg}')

    def _setup_job(self):
        os.makedirs(self.workdir, exist_ok=True)

        # set up the input file. This should always contain calling of the densf program, as per the SCM documentation
        with open(self.inputfile_path, "w+") as inpf:
            inpf.write("$AMSBIN/densf << eor\n")
            inpf.write(f"ADFFile {os.path.abspath(self.settings.ADFFile)}\n")
            inpf.write(f"GRID {self.settings.grid}\n")
            inpf.write("END\n")

            if len(self._mos) > 0:
                inpf.write("Orbitals SCF\n")
                for orb in self._mos:
                    inpf.write(f"    {orb.symmetry} {orb.index_in_symlabel + 1}\n")
                inpf.write("END\n")

            if len(self._sfos) > 0:
                inpf.write("Orbitals SFO\n")
                for orb in self._sfos:
                    inpf.write(f"    {orb.symmetry} {orb.index}\n")
                inpf.write("END\n")

            for line in self._extras:
                inpf.write(line + "\n")

            # cuboutput prefix is always the original run directory containing the adf.rkf file and includes the grid size
            outname = self.settings.grid if self.settings.grid.lower() in ['coarse', 'medium', 'fine'] else 'custom_grid'
            inpf.write(f"CUBOUTPUT {os.path.split(os.path.abspath(self.settings.ADFFile))[0]}/{outname}\n")
            inpf.write("eor\n")

        # the runfile should simply execute the input file.
        with open(self.runfile_path, "w+") as runf:
            runf.write("#!/bin/sh\n\n")
            runf.write("\n".join(self._preambles) + "\n\n")
            runf.write(f"sh {self.inputfile_path}\n")
            runf.write("\n".join(self._postambles))

        return True

    @property
    def output_cub_paths(self):
        """
        The output cube file paths that will be/were calculated by this job.
        """
        paths = []
        cuboutput = f"{os.path.split(os.path.abspath(self.settings.ADFFile))[0]}/{self.settings.grid}"

        for mo in self._mos:
            spin_part = "" if mo.spin == "AB" else f"_{mo.spin}"
            paths.append(f"{cuboutput}%SCF_{mo.symmetry}{spin_part}%{mo.index_in_symlabel + 1}.cub")

        for sfo in self._sfos:
            spin_part = "" if sfo.spin == "AB" else f"_{sfo.spin}"
            paths.append(f"{cuboutput}%SFO_{sfo.symmetry}{spin_part}%{sfo.index}.cub")

        for extra in self._extras:
            if extra == "Density SCF":
                paths.append(f"{cuboutput}%SCF%Density.cub")

            if extra.startswith('NCI'):
                paths.append(f'{cuboutput}%SCF%FitDenSigned.cub')
                paths.append(f'{cuboutput}%SCF%Fitdensity.cub')
                paths.append(f'{cuboutput}%SCF%FitNCI.cub')
                paths.append(f'{cuboutput}%SCF%FitRDG.cub')
                paths.append(f'{cuboutput}%SCF%FitRDGforNCI.cub')
                if 'both' in extra.lower():
                    paths.append(f'{cuboutput}%SCF%DenSigned.cub')
                    paths.append(f'{cuboutput}%SCF%Density.cub')
                    paths.append(f'{cuboutput}%SCF%NCI.cub')
                    paths.append(f'{cuboutput}%SCF%RDG.cub')
                    paths.append(f'{cuboutput}%SCF%RDGforNCI.cub')


        return paths

    def can_skip(self):
        return all(os.path.exists(path) for path in self.output_cub_paths)


if __name__ == "__main__":
    # import pyfmo

    # orbs = pyfmo.orbitals.Orbitals('/Users/yumanhordijk/PhD/MM2024/calculations/IRC/pi_beta_trans_TS1/pi_beta/pi_beta_trans/complex.00039/adf.rkf')
    # with DensfJob() as job:
    #     job.orbital(orbs.sfos['frag1(HOMO)'])

    with ADFFragmentJob() as job:
        ...



import pathlib as pl

import pytest
from tcutility import results


@pytest.fixture()
def ethane_res() -> results.Result:
    return results.read(pl.Path(__file__).parent / "fixtures" / "ethane_adf")


@pytest.fixture()
def unrestricted_res() -> results.Result:
    return results.read(pl.Path(__file__).parent / "fixtures" / "radical_addition_ts/geometry")


def test_energy_terms(ethane_res):
    print(ethane_res.keys())
    assert ethane_res.properties.energy.gibbs == pytest.approx(-522.38, abs=1e-2)
    assert ethane_res.properties.energy.enthalpy == pytest.approx(-507.64, abs=1e-2)
    assert ethane_res.properties.energy.nuclear_internal == pytest.approx(29.09, abs=1e-2)


def test_s2_restricted(ethane_res):
    assert ethane_res.properties.s2 == 0
    assert ethane_res.properties.s2_expected == 0
    assert ethane_res.properties.spin_contamination == 0


def test_s2_spinpol1(unrestricted_res):
    assert round(unrestricted_res.properties.s2, 5) == 0.75845
    assert unrestricted_res.properties.s2_expected == 0.75
    assert round(unrestricted_res.properties.spin_contamination, 4) == 0.0113


if __name__ == '__main__':
    pytest.main()



from scm import plams
import tcutility
from tcutility import log
from tcutility.job.generic import Job
import os
import numpy as np
from typing import List

j = os.path.join


class AMSJob(Job):
    '''
    This is the AMS base job which will serve as the parent class for ADFJob, DFTBJob and the future BANDJob.
    It holds all methods related to changing the settings at the AMS level. It also handles preparing the jobs, e.g. writing runfiles and inputs.
    '''
    def __str__(self):
        return f'{self._task}({self._functional}/{self._basis_set}), running in {os.path.join(os.path.abspath(self.rundir), self.name)}'

    def single_point(self):
        '''
        Set the task of the job to single point.
        '''
        self._task = 'SP'
        self.settings.input.ams.task = 'SinglePoint'

    def transition_state(self, distances: list = None, angles: list = None, dihedrals: list = None, ModeToFollow: int = 1):
        '''
        Set the task of the job to transition state search. Optionally you can give some TS coordinates to accelerate convergence.
        By default also calculates the normal modes after convergence.

        Args:
            distances: sequence of tuples or lists containing [atom_index1, atom_index2, factor]. Atom indices start at 1.
            angles: sequence of tuples or lists containing [atom_index1, atom_index2, atom_index3, factor]. Atom indices start at 1.
            dihedrals: sequence of tuples or lists containing [atom_index1, atom_index2, atom_index3, atom_index4, factor]. Atom indices start at 1.
            ModeToFollow: the vibrational mode to follow during optimization.
        '''
        self._task = 'TS'
        self.settings.input.ams.task = 'TransitionStateSearch'

        self.settings.input.ams.TransitionStateSearch.ModeToFollow = ModeToFollow

        if distances is not None:
            self.settings.input.ams.TransitionStateSearch.ReactionCoordinate.Distance = [" ".join([str(x) for x in dist]) for dist in distances]
        if angles is not None:
            self.settings.input.ams.TransitionStateSearch.ReactionCoordinate.Angle = [" ".join([str(x) for x in ang]) for ang in angles]
        if dihedrals is not None:
            self.settings.input.ams.TransitionStateSearch.ReactionCoordinate.Dihedral = [" ".join([str(x) for x in dihedral]) for dihedral in dihedrals]

        # for TS searches we quickly calculate the hessian with DFTB
        self.settings.input.ams.GeometryOptimization.InitialHessian.Type = 'CalculateWithFastEngine'
        self.vibrations(True)  # also calculate vibrations by default

    def optimization(self):
        '''
        Set the task of the job to geometry optimization. By default also calculates the normal modes after convergence.
        '''
        self._task = 'GO'
        self.settings.input.ams.task = 'GeometryOptimization'
        self.settings.input.ams.GeometryOptimization.InitialHessian.Type = 'CalculateWithFastEngine'
        self.vibrations(True)

    def IRC(self, direction: str = 'both', hess_file: str = None, step_size: float = 0.2, min_path_length: float = 0.1, max_points: int = 300):
        '''
        Set the task of the job to intrinsic reaction coordinate (IRC).

        Args:
            direction: the direction to take the first step into. By default it will be set to ``both``.
            hess_file: the path to a ``.rkf`` file to read the Hessian from. This is the ``adf.rkf`` file for ADF calculations.
                If set to ``None`` the Hessian will be calculated prior to starting the IRC calculation.
            step_size: the size of the step taken between each constrained optimization. By default it will be set to ``0.2`` :math:`a_0\\sqrt{Da}`.
            min_path_length: the length of the IRC path before switching to minimization. By default it will be set to ``0.1`` |angstrom|.
            max_points: the maximum number of IRC points in a direction. Be default it is set to ``300``.
        '''
        self._task = 'IRC'
        self.settings.input.ams.task = 'IRC'
        self.settings.input.ams.IRC.Direction = direction
        if hess_file:
            self.settings.input.ams.IRC.InitialHessian.File = hess_file
            self.settings.input.ams.IRC.InitialHessian.Type = 'FromFile'
        self.settings.input.ams.IRC.Step = step_size
        self.settings.input.ams.IRC.MinPathLength = min_path_length
        self.settings.input.ams.IRC.MaxPoints = max_points

        self.add_postscript(tcutility.job.postscripts.clean_workdir)
        self.add_postscript(tcutility.job.postscripts.write_converged_geoms)

    def PESScan(self, distances: list = None, angles: list = None, dihedrals: list = None, sumdists: list = None, difdists: list = None, npoints: int = 10):
        '''
        Set the task of the job to potential energy surface scan (PESScan).

        Args:
            distances: sequence of tuples or lists containing ``[atom_index1, atom_index2, start, end]``.
                Atom indices start at 1. Distances are given in |angstrom|.
            angles: sequence of tuples or lists containing ``[atom_index1, atom_index2, atom_index3, start, end]``.
                Atom indices start at 1. Angles are given in degrees
            dihedrals: sequence of tuples or lists containing ``[atom_index1, atom_index2, atom_index3, atom_index4, start, end]``.
                Atom indices start at 1. Angles are given in degrees
            sumdists: sequence of tuples or lists containing ``[atom_index1, atom_index2, atom_index3, atom_index4, start, end]``.
                Atom indices start at 1. Sum of distances is given in |angstrom|.
            difdists: sequence of tuples or lists containing ``[atom_index1, atom_index2, atom_index3, atom_index4, start, end]``.
                Atom indices start at 1. Difference of distances is given in |angstrom|.
            npoints: the number of PES points to optimize.

        .. note::
            Currently we only support generating settings for 1-dimensional PESScans.
            We will add support for N-dimensional PESScans later.
        '''
        self._task = 'PESScan'
        self.settings.input.ams.task = 'PESScan'
        self.settings.input.ams.PESScan.ScanCoordinate.nPoints = npoints
        if distances is not None:
            self.settings.input.ams.PESScan.ScanCoordinate.Distance = [" ".join([str(x) for x in dist]) for dist in distances]
        if angles is not None:
            self.settings.input.ams.PESScan.ScanCoordinate.Angle = [" ".join([str(x) for x in ang]) for ang in angles]
        if dihedrals is not None:
            self.settings.input.ams.PESScan.ScanCoordinate.Dihedral = [" ".join([str(x) for x in dihedral]) for dihedral in dihedrals]
        if sumdists is not None:
            self.settings.input.ams.PESScan.ScanCoordinate.SumDist = [" ".join([str(x) for x in dist]) for dist in sumdists]
        if difdists is not None:
            self.settings.input.ams.PESScan.ScanCoordinate.DifDist = [" ".join([str(x) for x in dist]) for dist in difdists]

        self.add_postscript(tcutility.job.postscripts.clean_workdir)
        self.add_postscript(tcutility.job.postscripts.write_converged_geoms)

    def vibrations(self, enable: bool = True, NegativeFrequenciesTolerance: float = 0.0):
        '''
        Set the calculation of vibrational modes.

        Args:
            enable: whether to calculate the vibrational modes.
            NegativeFrequenciesTolerance: the tolerance for negative modes.
                Modes with frequencies above this value will not be counted as imaginary.
                Use this option when you experience a lot of numerical noise.
        '''
        self.settings.input.ams.Properties.NormalModes = 'Yes' if enable else 'No'
        self.settings.input.ams.Properties.PESPointCharacter = 'Yes'

        self.settings.input.ams.PESPointCharacter.NegativeFrequenciesTolerance = NegativeFrequenciesTolerance
        self.settings.input.ams.NormalModes.ReScanFreqRange = '-10000000.0 10.0'

    def geometry_convergence(self, gradients: float = 1e-5, energy: float = 1e-5, step: float = 1e-2, stress: float = 5e-4):
        '''
        Set the convergence criteria for the geometry optimization.

        Args:
            gradients: the convergence criteria for the gradients during geometry optimizations. Defaults to ``1e-5``.
            energy: the convergence criteria for the energy during geometry optimizations. Defaults to ``1e-5``.
            step: the convergence criteria for the step-size during geometry optimizations. Defaults to ``1e-2``.
            stress: the convergence criteria for the stress-energy per atom during geometry optimizations. Defaults to ``5e-4``
        '''
        self.settings.input.ams.GeometryOptimization.Convergence.Gradients = gradients
        self.settings.input.ams.GeometryOptimization.Convergence.Energy = energy
        self.settings.input.ams.GeometryOptimization.Convergence.Step = step
        self.settings.input.ams.GeometryOptimization.Convergence.StressEnergyPerAtom = stress

    def charge(self, val: int):
        '''
        Set the charge of the system.
        '''
        self.settings.input.ams.System.Charge = val

    def electric_field(self, direction: List[float], magnitude: float = None):
        '''
        Set an electric field for this system.

        Args:
            direction: the vector with the direction and strength of the electric field.
            magnitude: if given, the direction will be normalized and magnitude will be used as the field strength.
        '''
        # if magnitude is given we normalize the direction vector and give it the correct length
        if magnitude is not None:
            direction = np.array(direction)/np.linalg.norm(direction) * magnitude

        ex, ey, ez = tuple(direction)
        self.settings.input.ams.System.ElectrostaticEmbedding.ElectricField = f'{ex} {ey} {ez}'

    def _setup_job(self):
        '''
        Set up the calculation. This will create the working directory and write the runscript and input file for ADF to use.
        '''
        os.makedirs(self.rundir, exist_ok=True)

        if os.path.exists(self.workdir):
            for file in os.listdir(self.workdir):
                p = os.path.join(self.workdir, file)
                if p.endswith('.rkf'):
                    os.remove(p)
                if 'ams.kid' in p:
                    os.remove(p)
                if p.startswith('t21.'):
                    os.remove(p)
                if p.startswith('t12.'):
                    os.remove(p)

        if not self._molecule and not self._molecule_path and 'atoms' not in self.settings.input.ams.system:
            log.error(f'You did not supply a molecule for this job. Call the {self.__class__.__name__}.molecule method to add one.')
            return

        if not self._molecule:
            self.settings.input.ams.system.GeometryFile = self._molecule_path

        # sometimes we have to check if a job is able to run or requires some special consideration
        # those types of checks should be defined in _check_job
        self._check_job()

        # we will use plams to write the input and runscript
        sett = self.settings.as_plams_settings()
        job = plams.AMSJob(name=self.name, molecule=self._molecule, settings=sett)

        os.makedirs(self.workdir, exist_ok=True)
        with open(self.inputfile_path, 'w+') as inpf:
            inpf.write(job.get_input())

        with open(self.runfile_path, 'w+') as runf:
            runf.write('#!/bin/sh\n\n')  # the shebang is not written by default by ADF
            runf.write('\n'.join(self._preambles) + '\n\n')
            runf.write(job.get_runscript())
            runf.write('\n'.join(self._postambles))

        # in case we are rerunning a calculation we need to remove ams.log
        if os.path.exists(j(self.workdir, 'ams.log')):
            os.remove(j(self.workdir, 'ams.log'))

        return True

    def _check_job(self):
        ...

    @property
    def output_mol_path(self):
        '''
        The default file path for output molecules when running ADF calculations. It will not be created for singlepoint calculations.
        '''
        return j(self.workdir, 'output.xyz')



import numpy as np
from tcutility import log
from typing import List


def naive_recursive(a: str, b: str) -> float:
    '''
    The naïve recursive algorithm to obtain the Levenshtein distance between two strings.
    We do not recommend using this algorithm as it is quite slow and faster alternatives exist.

    Args:
        a, b: strings to compare.

    Returns:
        The Levenshtein distance between the strings ``a`` and ``b``.

    .. seealso::
        :func:`wagner_fischer`
            A more efficient algorithm to obtain the Levenshtein distance (up to 25x faster).
    '''
    if len(b) == 0:
        return len(a)
    if len(a) == 0:
        return len(b)

    def tail(s):
        return s[1:]

    if a[0] == b[0]:
        return naive_recursive(tail(a), tail(b))

    d = 1 + min(naive_recursive(tail(a), b), naive_recursive(a, tail(b)), naive_recursive(tail(a), tail(b)))
    return d


def wagner_fischer(a: str, b: str, substitution_cost: float = 1, case_missmatch_cost: float = 1, insertion_cost: float = 1) -> float:
    '''
    Return the Levenshtein distance using the Wagner-Fischer algorithm.
    You can also change the penalty for various errors for this algorithm.
    By default, all types of errors incur a penalty of 1.

    Args:
        a, b: strings to compare.
        substitution_cost: the penalty for the erroneous substitution of a character.
        case_missmatch_cost: the penalty for miss-matching the case of a character.
        insertion_cost: the cost for the erroneous insertion or deletion of a character.
    
    Returns:
        The Levenshtein distance between the strings ``a`` and ``b``.

    Example:
        .. code-block:: python

            >>> wagner_fischer('kitten', 'sitting')
            3

    .. seealso::
        :func:`naive_recursive`
            An alternative (and slower) algorithm to obtain the Levenshtein distance.
    '''
    # initialize an empty matrix
    d = np.zeros((len(a)+1, len(b)+1)).astype(float)

    # the top row and left column are always the same
    for i in range(len(a)):
        d[i+1, 0] = i+1
    for i in range(len(b)):
        d[0, i+1] = i+1

    # iteratively build the matrix
    for i in range(1, len(a)+1):
        for j in range(1, len(b)+1):
            # if the two characters are the same the cost is 0
            if a[i-1] == b[j-1]:
                cost = 0
            # case miss-match
            elif a[i-1].lower() == b[j-1].lower():
                cost = case_missmatch_cost
            # substitution
            else:
                cost = substitution_cost

            # here determine if the substitution/miss-match cost is greater than the insertion or deletion cost
            d[i, j] = min(d[i-1, j] + insertion_cost, d[i, j-1] + insertion_cost, d[i-1, j-1] + cost)

    # return the bottom-right element, this is the final edit-distance
    return d[-1, -1]


def get_closest(a: str, others: List[str], compare_func=wagner_fischer, ignore_case: bool = False, ignore_chars: str = '', maximum_distance: int = None, **kwargs) -> List[str]:
    '''
    Return strings that are similar to an input string using the Levenshtein distance.

    Args:
        a: the string to compare the rest to.
        others: a collection of strings to compare to a. The returned strings will be taken from this collection.
        compare_func: the function to use to compare the strings. Defaults to the efficient :func:`wagner_fischer` algorithm.
        ignore_case: whether the case of the strings is taken into account. If enabled, all strings are turned to lower-case before comparison.
        ignore_chars: a strings specifying characters that should be ignored.
        maximum_distance: the maximum Levenshtein distance to allow. 
            If it is lower than the lowest distance for the collection of strings, we return the strings with the lowest distance. 
            If set to ``None`` we return the lowest distance strings.

    Returns:
        A collection of strings that have a Levenshtein distance to ``a`` below ``maximum_distance`` 
        or have the lowest distance to ``a`` if all strings have a distance greater than ``maximum_distance``.
        If the lowest distance is ``0``, return an empty list instead.

    Example:
        .. code-block:: python

            >>> closest = get_closest('kitten', ['mitten', 'bitten', 'sitting'])
            >>> print(closest)
            ['mitten', 'bitten']
    '''
    if ignore_case:
        a = a.lower()

    for char in ignore_chars:
        a = a.replace(char, '')

    dists = []
    for other in others:
        if ignore_case:
            other = other.lower()

        for char in ignore_chars:
            other = other.replace(char, '')

        dists.append(compare_func(a, other, **kwargs))

    if maximum_distance is None:
        maximum_distance = -1
    lowest_strs = [other for dist, other in zip(dists, others) if 0 < dist <= max(maximum_distance, min(dists))]
    return lowest_strs


def make_suggestion(a: str, others: List[str], **kwargs):
    '''
    Print a warning that gives suggestions for strings that are close to a given string.

    Example:
        .. code-block:: python

            >>> make_suggestion('kitten', ['mitten', 'bitten', 'sitting'])
            [2024/01/30 15:26:35] [WARNING](main): Could not find "kitten". Did you mean mitten or bitten?

    .. seealso::
        :func:`get_closest` for a description of the function arguments.
    '''
    closest = get_closest(a, others, **kwargs)

    if len(closest) > 1:
        closest_string = " or ".join([", ".join(closest[:-1]), closest[-1]])
    else:
        closest_string = closest[0]

    # write a warning message
    log.warn(f'Could not find "{a}". Did you mean {closest_string}?', caller_level=3)


def check(a: str, others: List[str], caller_level=2, **kwargs):
    closest = get_closest(a, others, **kwargs)
    if len(closest) == 0:
        return

    if len(closest) > 1:
        closest_string = " or ".join([", ".join(closest[:-1]), closest[-1]])
    else:
        closest_string = closest[0]

    # write a warning message
    msg = f"({log.caller_name(caller_level)}): " + f'Could not find "{a}". Did you mean {closest_string}?'
    raise ValueError(msg)



from pprint import pprint
from typing import List

import click
from tcutility import results


@click.command()
@click.option("-s", "--status", is_flag=True, help="Shortcut to only print the status of the calculation.")
@click.option("-p", "--properties", is_flag=True, help="Shortcut to only print calculated properties for the calculation.")
@click.argument("workdir")
@click.argument("keys", nargs=-1)
def read_results(status: bool, properties: bool, workdir: str, keys: List[str]):
    """Read results from a calculation."""
    res = results.read(workdir)

    if status:
        print(res.status.name)  # type: ignore # status is a str
        return

    if properties:
        pprint(res.properties)
        return

    if len(keys) > 0:
        ret = {k: res.get_multi_key(k) for k in keys}
        if len(ret) == 1:
            print(list(ret.values())[0])
            return
        pprint(ret)
        return

    pprint(res)



import os


# pwd = os.path.getcwd()
# print(pwd)

candidates = ['crest_conformers.xyz', 'crest_rotamers.xyz', 'ensemble/final_ensemble.xyz']
out_dirs = ['conformers', 'rotamers', 'ensembles']
for out_dir, candidate in zip(out_dirs, candidates):
	if not os.path.exists(candidate):
		continue

	os.makedirs(out_dir, exist_ok=True)
	with open(candidate) as xyz:
		lines = xyz.readlines()
		natoms = int(lines[0].strip())
		nmols = len(lines)//(natoms + 2)
		mol_lines = [lines[(natoms+2)*i:(natoms+2)*(i+1)] for i in range(nmols)]

	[[print(line.strip()) for line in lines] for lines in mol_lines]


# [print(x) for x in geometry.get_rotmat.__globals__]
# print(geometry.get_rotmat.__globals__['__file__'])



from . import postscripts  # noqa: F401
from .generic import Job  # noqa: F401
from .ams import AMSJob  # noqa: F401
from .adf import ADFJob, ADFFragmentJob, DensfJob  # noqa: F401
from .dftb import DFTBJob  # noqa: F401
from .crest import CRESTJob, QCGJob  # noqa: F401
from .xtb import XTBJob  # noqa: F401
from .orca import ORCAJob  # noqa: F401
from .nmr import NMRJob  # noqa: F401



if __name__ == '__main__':
    import os
    import glob

    if len(glob.glob('IRC*converged*.rkf')) > 0:
        if os.path.exists('TS.rkf'):
            os.remove('TS.rkf')

    for f in glob.glob('IRC*converged*.rkf'):
        os.remove(f)

    for f in glob.glob('t21*'):
        os.remove(f)

    for f in glob.glob('t12*'):
        os.remove(f)

    if os.path.exists('CreateAtoms.out'):
        os.remove('CreateAtoms.out')



from typing import Dict, List

import numpy as np
from scm.plams import KFReader

from tcutility import constants, ensure_list
from tcutility.results import Result, cache
from tcutility.tc_typing import arrays

# ------------------------------------------------------------- #
# ------------------- Calculation Settings -------------------- #
# ------------------------------------------------------------- #


def get_calc_settings(info: Result) -> Result:
    """Function to read calculation settings for an ADF calculation.

    Args:
        info: Result object containing ADF calculation settings.

    Returns:
        :Result object containing properties from the ADF calculation:

            - **task (str)** – the task that was set for the calculation.
            - **relativistic (bool)** – whether or not relativistic effects were enabled.
            - **relativistic_type (str)** – the name of the relativistic approximation used.
            - **unrestricted_sfos (bool)** – whether or not SFOs are treated in an unrestricted manner.
            - **unrestricted_mos (bool)** – whether or not MOs are treated in an unrestricted manner.
            - **symmetry.group (str)** – the symmetry group selected for the molecule.
            - **symmetry.labels (list[str])** – the symmetry labels associated with the symmetry group.
            - **used_regions (bool)** – whether regions were used in the calculation.
            - **charge (int)** - the charge of the system.
            - **spin_polarization (int)** - the spin-polarization of the system.
            - **multiplicity (int)** - the multiplicity of the system. This is equal to 2|S|+1 for spin-polarization S.
    """
    assert info.engine == "adf", f"This function reads ADF data, not {info.engine} data"

    ret = Result()

    # set the calculation task at a higher level
    ret.task = info.input.task

    # the VibrationalAnalysis task does not produce adf.rkf files
    # in that case we end the reading here, since the following
    # variables do not apply
    if ret.task.lower() == "vibrationalanalysis":
        return ret

    reader_adf = cache.get(info.files["adf.rkf"])

    relativistic_type_map = {
        0: "None",
        1: "scalar Pauli",
        3: "scalar ZORA",  # scalar ZORA + MAPA
        4: "scalar ZORA + full pot.",
        5: "scalar ZORA + APA",
        6: "scalar X2C + MAPA",
        7: "scalar X2C ZORA + MAPA",
        11: "spin-orbit Pauli",
        13: "spin-orbit ZORA",  # spin-orbit ZORA + MAPA
        14: "spin-orbit ZORA + full pot.",
        15: "spin-orbit ZORA + APA",
        16: "spin-orbit X2C + MAPA",
        17: "spin-orbit X2C ZORA + MAPA",
    }
    # determine if calculation used relativistic corrections
    # if it did, variable 'escale' will be present in 'SFOs'
    # if it didnt, only variable 'energy' will be present
    ret.relativistic = ("SFOs", "escale") in reader_adf
    ret.relativistic_type = relativistic_type_map[reader_adf.read("General", "ioprel")]

    # determine if MOs are unrestricted or not
    # general, nspin is 1 for restricted and 2 for unrestricted calculations
    ret.unrestricted_mos = reader_adf.read("General", "nspin") == 2

    # determine if SFOs are unrestricted or not
    ret.unrestricted_sfos = reader_adf.read("General", "nspinf") == 2

    # get the symmetry group
    ret.symmetry.group = reader_adf.read("Geometry", "grouplabel").strip()

    # get the symmetry labels
    ret.symmetry.labels = reader_adf.read("Symmetry", "symlab").strip().split()

    # determine if the calculation used regions or not
    frag_order = reader_adf.read("Geometry", "fragment and atomtype index")
    frag_order = frag_order[: len(frag_order) // 2]
    ret.used_regions = max(frag_order) != len(frag_order)

    ret.charge = reader_adf.read("Molecule", "Charge")

    ret.spin_polarization = 0
    if ret.unrestricted_mos:
        nalpha = 0
        nbeta = 0
        for label in ret.symmetry.labels:
            nalpha += sum(ensure_list(reader_adf.read(label, "froc_A")))
            nbeta += sum(ensure_list(reader_adf.read(label, "froc_B")))
        ret.spin_polarization = nalpha - nbeta
    ret.multiplicity = 2 * ret.spin_polarization + 1

    return ret


# ------------------------------------------------------------- #
# ------------------------ Properties ------------------------- #
# ------------------------------------------------------------- #


# ----------------------- VDD charges ------------------------- #


def _read_vdd_charges(kf_reader: KFReader) -> arrays.Array1D[np.float64]:
    """Returns the VDD charges from the KFReader object."""
    vdd_scf: List[float] = ensure_list(kf_reader.read("Properties", "AtomCharge_SCF Voronoi"))  # type: ignore since plams does not include typing for KFReader. List[float] is returned
    vdd_ini: List[float] = ensure_list(kf_reader.read("Properties", "AtomCharge_initial Voronoi"))  # type: ignore since plams does not include typing for KFReader. List[float] is returned

    # VDD charges are scf - initial charges. Note, these are in units of electrons while most often these are denoted in mili-electrons
    return np.array([float((scf - ini)) for scf, ini in zip(vdd_scf, vdd_ini)])


def _read_vdd_charges_initial(kf_reader: KFReader) -> arrays.Array1D[np.float64]:
    """Returns the initial VDD charges from the KFReader object."""
    vdd_ini: List[float] = ensure_list(kf_reader.read("Properties", "AtomCharge_initial Voronoi"))  # type: ignore since plams does not include typing for KFReader. List[float] is returned
    return np.array(vdd_ini)


def _read_vdd_charges_SCF(kf_reader: KFReader) -> arrays.Array1D[np.float64]:
    """Returns the SCF VDD charges from the KFReader object."""
    vdd_scf: List[float] = ensure_list(kf_reader.read("Properties", "AtomCharge_SCF Voronoi"))  # type: ignore since plams does not include typing for KFReader. List[float] is returned
    return np.array(vdd_scf)


def _get_vdd_charges_per_irrep(results_type: KFReader) -> Dict[str, arrays.Array1D[np.float64]]:
    """Extracts the Voronoi Deformation Density charges from the fragment calculation sorted per irrep."""
    symlabels = str(results_type.read("Symmetry", "symlab")).split()  # split on whitespace

    # If there is only one irrep, there is no irrep decomposition
    if len(symlabels) == 1:
        return {}

    vdd_irrep = np.array(results_type.read("Properties", "Voronoi chrg per irrep"))
    n_atoms = int(results_type.read("Molecule", "nAtoms"))  # type: ignore since no static typing. Returns an int

    # NOTE: apparently, the irrep charges are minus the total VDD charges. That's why the minus sign in the second line below
    vdd_irrep = vdd_irrep[-len(symlabels) * n_atoms :]  # vdd_irrep = vdd_irrep.reshape((n_headers, len(symlabels), n_atoms)) # NOQA E203
    vdd_per_irrep = {irrep: -vdd_irrep[i * n_atoms : (i + 1) * n_atoms] for i, irrep in enumerate(symlabels)}  # NOQA E203
    return vdd_per_irrep


# ----------------------- Vibrations ------------------------- #


def _read_vibrations(reader: cache.TrackKFReader) -> Result:
    ret = Result()
    ret.number_of_modes = reader.read("Vibrations", "nNormalModes")
    ret.frequencies = ensure_list(reader.read("Vibrations", "Frequencies[cm-1]"))
    if ("Vibrations", "Intensities[km/mol]") in reader:
        ret.intensities = ensure_list(reader.read("Vibrations", "Intensities[km/mol]"))
    ret.number_of_imag_modes = len([freq for freq in ret.frequencies if freq < 0])
    ret.character = "minimum" if ret.number_of_imag_modes == 0 else "transitionstate"
    ret.modes = []
    for i in range(ret.number_of_modes):
        ret.modes.append(reader.read("Vibrations", f"NoWeightNormalMode({i+1})"))
    return ret


def get_properties(info: Result) -> Result:
    """Function to get properties from an ADF calculation.

    Args:
        info: Result object containing ADF properties.

    Returns:
        :Result object containing properties from the ADF calculation:

            - **energy.bond (float)** – bonding energy (|kcal/mol|).
            - **energy.elstat.total (float)** – total electrostatic potential (|kcal/mol|).
            - **energy.elstat.Vee (float)** – electron-electron repulsive term of the electrostatic potential (|kcal/mol|).
            - **energy.elstat.Ven (float)** – electron-nucleus attractive term of the electrostatic potential (|kcal/mol|).
            - **energy.elstat.Vnn (float)** – nucleus-nucleus repulsive term of the electrostatic potential (|kcal/mol|).
            - **energy.orbint.total (float)** – total orbital interaction energy containing contributions from each symmetry label and correction energy(|kcal/mol|).
            - **energy.orbint.{symmetry label} (float)** – orbital interaction energy from a specific symmetry label (|kcal/mol|).
            - **energy.orbint.correction (float)** - orbital interaction correction energy, the difference between the total and the sum of the symmetrized interaction energies (|kcal/mol|)
            - **energy.pauli.total (float)** – total Pauli repulsion energy (|kcal/mol|).
            - **energy.dispersion (float)** – total dispersion energy (|kcal/mol|).
            - **energy.gibbs (float)** – Gibb's free energy (|kcal/mol|). Only populated if vibrational modes were calculated.
            - **energy.enthalpy (float)** – enthalpy (|kcal/mol|). Only populated if vibrational modes were calculated.
            - **energy.nuclear_internal (float)** – nuclear internal energy (|kcal/mol|). Only populated if vibrational modes were calculated.
            - **vibrations.number_of_modes (int)** – number of vibrational modes for this molecule, 3N-5 for non-linear molecules and 3N-6 for linear molecules, where N is the number of atoms.
            - **vibrations.number_of_imag_modes (int)** – number of imaginary vibrational modes for this molecule.
            - **vibrations.frequencies (float)** – vibrational frequencies associated with the vibrational modes, sorted from low to high (|cm-1|).
            - **vibrations.intensities (float)** – vibrational intensities associated with the vibrational modes (|km/mol|).
            - **vibrations.modes (list[float])** – list of vibrational modes sorted from low frequency to high frequency.
            - **vibrations.character (str)** – Character of the molecule based on the number of imaginary vibrational modes. Can be "minimum" or "transition state".
            - **vdd.charges (nparray[float] (1D))** - 1D array of Voronoi Deformation Denisty (VDD) charges in [electrons], being the difference between the final (SCF) and initial VDD charges.
            - **vdd.charges.{symmetry label} (nparray[float] (1D))** - 1D array of Voronoi Deformation Denisty (VDD) charges in [electrons] per irrep.
            - **s2** - expectation value of the :math:`S^2` operator.
            - **s2_expected** - ideal expectation value of the :math:`S^2` operator. For restricted calculations this should always equal ``s2``.
            - **spin_contamination** - the amount of spin-contamination observed in this calculation. It is equal to (s2 - s2_expected) / (s2_expected). Ideally this value should be below 0.1.
            - **dipole_vector** - the dipole moment vector.
            - **dipole_moment** - the magnitude of the dipole moment vector.
            - **quadrupole_moment** - the quadrupole moment vector.
            - **dens_at_atom** - the electron density at each atom.
    """

    assert info.engine == "adf", f"This function reads ADF data, not {info.engine} data"

    ret = Result()

    if info.adf.task.lower() == "vibrationalanalysis":
        reader_ams = cache.get(info.files["ams.rkf"])
        ret.vibrations = _read_vibrations(reader_ams)
        return ret

    reader_adf = cache.get(info.files["adf.rkf"])

    # read energies (given in Ha in rkf files)
    ret.energy.bond = reader_adf.read("Energy", "Bond Energy") * constants.HA2KCALMOL

    # total electrostatic potential
    ret.energy.elstat.total = reader_adf.read("Energy", "elstat") * constants.HA2KCALMOL

    # we can further decompose elstat if it was enabled
    if info.files.out:
        with open(info.files.out) as output:
            lines = output.readlines()

        skip_next = -1
        for line in lines:
            if "Electrostatic Interaction Energies" in line:
                skip_next = 4
                continue
            if skip_next == 0:
                f1, f2, Vee, Ven, Vnn, total = line.strip().split()
                ret.energy.elstat.Vee = float(Vee) * constants.HA2KCALMOL
                ret.energy.elstat.Ven = float(Ven) * constants.HA2KCALMOL
                ret.energy.elstat.Vnn = float(Vnn) * constants.HA2KCALMOL
            skip_next -= 1


    # print(info.files)

    # read the total orbital interaction energy
    ret.energy.orbint.total = reader_adf.read("Energy", "Orb.Int. Total") * constants.HA2KCALMOL

    # to calculate the orbital interaction term:
    # the difference between the total and the sum of the symmetrized interaction energies should be calculated
    # therefore the correction is first set equal to the total orbital interaction.
    ret.energy.orbint.correction = ret.energy.orbint.total

    # looping over every symlabel, to get the energy per symmetry label
    for symlabel in info.adf.symmetry.labels:
        symlabel = symlabel.split(":")[0]
        ret.energy.orbint[symlabel] = reader_adf.read("Energy", f"Orb.Int. {symlabel}") * constants.HA2KCALMOL

        # the energy per symmetry label is abstracted from the "total orbital interaction"
        # obtaining the correction to the orbital interaction term
        ret.energy.orbint.correction -= ret.energy.orbint[symlabel]

    ret.energy.pauli.total = reader_adf.read("Energy", "Pauli Total") * constants.HA2KCALMOL
    ret.energy.dispersion = reader_adf.read("Energy", "Dispersion Energy") * constants.HA2KCALMOL

    if ("Thermodynamics", "Gibbs free Energy") in reader_adf:
        ret.energy.gibbs = reader_adf.read("Thermodynamics", "Gibbs free Energy") * constants.HA2KCALMOL
        ret.energy.enthalpy = reader_adf.read("Thermodynamics", "Enthalpy") * constants.HA2KCALMOL
        ret.energy.nuclear_internal = reader_adf.read("Thermodynamics", "Internal Energy total") * constants.HA2KCALMOL

    # vibrational information
    if ("Vibrations", "nNormalModes") in reader_adf:
        ret.vibrations = _read_vibrations(reader_adf)

    # read the Voronoi Deformation Charges Deformation (VDD) before and after SCF convergence (being "inital" and "SCF")
    try:
        ret.vdd.charges = _read_vdd_charges(reader_adf)
        ret.vdd.charges_initial = _read_vdd_charges_initial(reader_adf)
        ret.vdd.charges_SCF = _read_vdd_charges_SCF(reader_adf)
        ret.vdd.update(_get_vdd_charges_per_irrep(reader_adf))
    except KeyError:
        pass

    # read spin-squared operator info
    # the total spin
    S = info.adf.spin_polarization * 1 / 2
    ret.s2_expected = S * (S + 1)
    # this is the real expectation value
    if ("Properties", "S2calc") in reader_adf:
        ret.s2 = reader_adf.read("Properties", "S2calc")
    else:
        ret.s2 = 0

    # calculate the contamination
    # if S is 0 then we will get a divide by zero error, but spin-contamination should be 0
    if S != 0:
        ret.spin_contamination = (ret.s2 - ret.s2_expected) / (ret.s2_expected)
    else:
        ret.spin_contamination = 0

    ret.dipole_vector = reader_adf.read("Properties", "Dipole")
    ret.dipole_moment = np.linalg.norm(ret.dipole_vector)
    ret.quadrupole_moment = reader_adf.read("Properties", "Quadrupole")
    ret.dens_at_atom = ensure_list(reader_adf.read("Properties", "Electron Density at Nuclei"))

    return ret


def get_level_of_theory(info: Result) -> Result:
    """Function to get the level-of-theory from an input-file.

    Args:
        inp_path: Path to the input file. Can be a .run or .in file create for AMS

    Returns:
        :Dictionary containing information about the level-of-theory:

            - **summary (str)** - a summary string that gives the level-of-theory in a human-readable format.
            - **xc.functional (str)** - XC-functional used in the calculation.
            - **xc.category (str)** - category the XC-functional belongs to. E.g. GGA, MetaHybrid, etc ...
            - **xc.dispersion (str)** - the dispersion correction method used during the calculation.
            - **xc.summary (str)** - a summary string that gives the XC-functional information in a human-readable format.
            - **xc.empiricalScaling (str)** - which empirical scaling parameter was used. Useful for MP2 calculations.
            - **basis.type (str)** - the size/type of the basis-set.
            - **basis.core (str)** - the size of the frozen-core approximation.
            - **quality (str)** - the numerical quality setting.
    """
    sett = info.input
    ret = Result()
    # print(json.dumps(sett, indent=4))
    xc_categories = ["GGA", "LDA", "MetaGGA", "MetaHybrid", "Model", "LibXC", "DoubleHybrid", "Hybrid", "MP2", "HartreeFock"]
    ret.xc.functional = "VWN"
    ret.xc.category = "LDA"
    for cat in xc_categories:
        if cat.lower() in [key.lower() for key in sett.adf.xc]:
            ret.xc.functional = sett.adf.xc[cat]
            ret.xc.category = cat

    ret.basis.type = sett.adf.basis.type
    ret.basis.core = sett.adf.basis.core
    ret.quality = sett.adf.NumericalQuality or "Normal"

    ret.xc.dispersion = None
    if "dispersion" in [key.lower() for key in sett.adf.xc]:
        ret.xc.dispersion = " ".join(sett.adf.xc.dispersion.split())

    # the empirical scaling value is used for MP2 calculations
    ret.xc.empirical_scaling = None
    if "empiricalscaling" in [key.lower() for key in sett.adf.xc]:
        ret.xc.empiricalscaling = sett.adf.xc.empiricalscaling

    # MP2 and HF are a little bit different from the usual xc categories. They are not key-value pairs but only the key
    # we start building the ret.xc.summary string here already. This will contain the human-readable functional name
    if ret.xc.category == "MP2":
        ret.xc.summary = "MP2"
        if ret.xc.empiricalscaling:
            ret.xc.summary += f"-{ret.xc.empiricalscaling}"
    elif ret.xc.category == "HartreeFock":
        ret.xc.summary = "HF"
    else:
        ret.xc.summary = ret.xc.functional

    # If dispersion was used, we want to add it to the ret.xc.summary
    if ret.xc.dispersion:
        if ret.xc.dispersion.lower() == "grimme3":
            ret.xc.summary += "-D3"
        if ret.xc.dispersion.lower() == "grimme3 bjdamp":
            ret.xc.summary += "-D3(BJ)"
        if ret.xc.dispersion.lower() == "grimme4":
            ret.xc.summary += "-D4"
        if ret.xc.dispersion.lower() == "ddsc":
            ret.xc.summary += "-dDsC"
        if ret.xc.dispersion.lower() == "uff":
            ret.xc.summary += "-dUFF"
        if ret.xc.dispersion.lower() == "mbd":
            ret.xc.summary += "-MBD@rsSC"
        if ret.xc.dispersion.lower() == "default":
            ret.xc.summary += "-D"

    # ret.summary is simply the ret.xc.summary plus the basis set type
    ret.summary = f"{ret.xc.summary}/{ret.basis.type}"
    return ret



import pathlib as pl

import numpy as np
import pytest
from scm.plams import KFReader
from tcutility import results
from tcutility.analysis.vdd import charge, manager
from tcutility.results import ams

# VDDCharge, manager.VDDChargeManager, get_vdd_charges, _get_fragment_indices_from_input_order

current_dir = pl.Path(__file__).parent
R_TOLERANCE = 1e-5  # relative tolerance for floating point comparisons


@pytest.fixture
def kfreader_fa_disordered_fragindices_cs():
    # replace 'your_file' with the path to your file
    return KFReader(str(current_dir / "fixtures" / "VDD" / "fa_acid_amide_cs" / "fa_acid_amide_cs.adf.rkf"))


@pytest.fixture
def vdd_manager_fa_disordered_fragindices_cs():
    fa_results = results.read(current_dir / "fixtures" / "VDD" / "fa_acid_amide_cs")
    return manager.create_vdd_charge_manager(fa_results)


@pytest.fixture
def kfreader_fa_ordered_fragindices_cs():
    return KFReader(str(current_dir / "fixtures" / "VDD" / "fa_squaramide_se_cs" / "fa_squaramide_se_cs.adf.rkf"))


@pytest.fixture
def vdd_manager_fa_ordered_fragindices_cs():
    fa_results = results.read(current_dir / "fixtures" / "VDD" / "fa_squaramide_se_cs")
    return manager.create_vdd_charge_manager(fa_results)


@pytest.fixture
def kfreader_fa_nosym():
    return KFReader(str(current_dir / "fixtures" / "VDD" / "fa_donor_acceptor_nosym" / "fa_donor_acceptor_nosym.adf.rkf"))


@pytest.fixture
def vdd_manager_fa_nosym():
    fa_results = results.read(current_dir / "fixtures" / "VDD" / "fa_donor_acceptor_nosym")
    return manager.create_vdd_charge_manager(fa_results)


@pytest.fixture
def kfreader_geo_nosym():
    return KFReader(str(current_dir / "fixtures" / "VDD" / "geo_nosym" / "geo_nosym.adf.rkf"))


@pytest.fixture
def vdd_manager_geo_nosym():
    geo_results = results.read(current_dir / "fixtures" / "VDD" / "geo_nosym" / "geo_nosym.adf.rkf")
    return manager.create_vdd_charge_manager(geo_results)


# ----------------------------------------------------------
# ------------------ get_fragment_indices ------------------
# ----------------------------------------------------------


def test_get_fragment_indices_from_input_order_disordered(kfreader_fa_disordered_fragindices_cs):
    frag_indices = ams._get_fragment_indices_from_input_order(kfreader_fa_disordered_fragindices_cs)
    assert np.array_equal(frag_indices, np.array([1, 1, 1, 2, 2, 2, 1, 1, 2, 2, 2]))


def test_get_fragment_indices_from_input_order_ordered(kfreader_fa_ordered_fragindices_cs):
    frag_indices = ams._get_fragment_indices_from_input_order(kfreader_fa_ordered_fragindices_cs)
    assert np.array_equal(frag_indices, np.array([1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2]))


# ----------------------------------------------------------
# --------------- manager.VDDChargeManager -----------------
# ----------------------------------------------------------


def test_change_unit(vdd_manager_fa_ordered_fragindices_cs: manager.VDDChargeManager):
    manager = vdd_manager_fa_ordered_fragindices_cs
    initial_charges = manager.get_vdd_charges("e")
    manager.change_unit("e")
    manager.change_unit("me")
    manager.change_unit("me")
    manager.change_unit("e")
    manager.change_unit("me")
    second_charges = manager.get_vdd_charges("e")
    other_unit_charges = manager.get_vdd_charges("me")

    assert np.allclose([charge.charge for charge in initial_charges["vdd"]], [charge.charge for charge in second_charges["vdd"]]), "Charges are not the same after changing units"
    assert np.allclose([charge.charge for charge in initial_charges["vdd"]], [charge.charge / 1000 for charge in other_unit_charges["vdd"]])


def test_get_vdd_charges_irrep_cs(vdd_manager_fa_ordered_fragindices_cs: manager.VDDChargeManager):
    vdd_charges = vdd_manager_fa_ordered_fragindices_cs.get_vdd_charges("e")

    assert len(vdd_charges) == 5, "More than three irreps found in the returned dict"
    assert "vdd" in vdd_charges, "'vdd' not found in the returned dict"
    assert "AA" in vdd_charges, "'AA' not found in the returned dict"
    assert "AAA" in vdd_charges, "'AAA' not found in the returned dict"


def test_get_vdd_charges_irrep_nosym(vdd_manager_fa_nosym: manager.VDDChargeManager):
    vdd_charges = vdd_manager_fa_nosym.get_vdd_charges("e")

    assert "vdd" in vdd_charges, "'vdd' not found in the returned dict"
    assert len(vdd_charges) == 3, "More than one irrep found in the returned dict"


def test_get_vdd_charges_charge_values_fa_cs(vdd_manager_fa_ordered_fragindices_cs: manager.VDDChargeManager):
    vdd_charges = vdd_manager_fa_ordered_fragindices_cs.get_vdd_charges("e")

    # The "/1000" is because the charges are in [electrons] in the kf file but it is more readable to use [millielectrons] here
    test_charge_1 = charge.VDDCharge(atom_index=1, atom_symbol="C", charge=-39.75281 / 1000, frag_index=1)
    test_charge_2 = charge.VDDCharge(atom_index=2, atom_symbol="Se", charge=-164.520 / 1000, frag_index=1)
    test_charge_3 = charge.VDDCharge(atom_index=8, atom_symbol="H", charge=53.397 / 1000, frag_index=2)
    assert np.isclose(vdd_charges["vdd"][0].charge, test_charge_1.charge), "Charge values are not as expected"
    assert np.isclose(vdd_charges["vdd"][1].charge, test_charge_2.charge), "Charge values are not as expected"
    assert np.isclose(vdd_charges["vdd"][7].charge, test_charge_3.charge), "Charge values are not as expected"
    assert np.isclose(np.sum([charge.charge for charge in vdd_charges["vdd"]]), 0.0, atol=1e-1), "Total charge is not zero"


def test_get_vdd_charges_charge_values_geo_nosym(vdd_manager_fa_nosym: manager.VDDChargeManager):
    vdd_charges = vdd_manager_fa_nosym.get_vdd_charges("e")

    # The "/1000" is because the charges are in [electrons] in the kf file but it is more readable to use [millielectrons] here
    test_charge_1 = charge.VDDCharge(atom_index=3, atom_symbol="C", charge=-0.328746 / 1000, frag_index=1)
    test_charge_2 = charge.VDDCharge(atom_index=4, atom_symbol="C", charge=-0.40333031 / 1000, frag_index=1)
    test_charge_3 = charge.VDDCharge(atom_index=12, atom_symbol="O", charge=-1.210724 / 1000, frag_index=2)
    assert np.isclose(vdd_charges["vdd"][2].charge, test_charge_1.charge), "Charge values are not as expected"
    assert np.isclose(vdd_charges["vdd"][3].charge, test_charge_2.charge), "Charge values are not as expected"
    assert np.isclose(vdd_charges["vdd"][11].charge, test_charge_3.charge), "Charge values are not as expected"
    assert np.isclose(np.sum([charge.charge for charge in vdd_charges["vdd"]]), 0.0, atol=1e-1), "Total charge is not zero"


def test_get_vdd_charges_irrep_atom_sum_cs(vdd_manager_fa_ordered_fragindices_cs: manager.VDDChargeManager):
    vdd_charges = vdd_manager_fa_ordered_fragindices_cs.get_vdd_charges("e")

    vdd_charge = vdd_charges["vdd"][0].charge
    aa_charge = vdd_charges["AA"][0].charge
    aaa_charge = vdd_charges["AAA"][0].charge
    assert np.isclose(vdd_charge, aa_charge + aaa_charge, rtol=R_TOLERANCE), "Total atom charge is not the sum of the irreps"


def test_summed_charges(vdd_manager_fa_ordered_fragindices_cs: manager.VDDChargeManager):
    summed_charges = vdd_manager_fa_ordered_fragindices_cs.get_summed_vdd_charges()  # in unit [me]

    assert len(summed_charges.keys()) == 5, "More than three irreps found in the summed charges"
    assert "AA" in summed_charges, "'AA' not found in the summed charges"
    assert "AAA" in summed_charges, "'AAA' not found in the summed charges"
    assert "vdd" in summed_charges, "'vdd' not found in the summed charges"

    frag1_total = summed_charges["vdd"]["1"]
    frag2_total = summed_charges["vdd"]["2"]
    frag1_aa = summed_charges["AA"]["1"]
    frag1_aaa = summed_charges["AAA"]["1"]
    assert np.isclose(frag1_total, -357.106, rtol=1e-5), "Summed charges total are not as expected"
    assert np.isclose(frag1_aa, 163.7488, rtol=1e-5), "Summed charges AA are not as expected"
    assert np.isclose(frag1_aaa, -520.8548, rtol=1e-5), "Summed charges AAA are not as expected"

    assert np.isclose(frag1_total, frag1_aa + frag1_aaa, atol=1e-2), "Total charge is not the sum of the irreps"
    assert np.isclose(frag1_total + frag2_total, 0, atol=1e-1), "Total charge is not the sum of the irreps"


def test_summed_charges_nosym(vdd_manager_fa_nosym: manager.VDDChargeManager):
    summed_charges = vdd_manager_fa_nosym.get_summed_vdd_charges()

    assert len(summed_charges.keys()) == 3, "Irreps found in the summed charges for nosym calculation"



from tcutility import formula


def test_single_molecule():
    assert formula.molecule("F-", "html") == "F<sup>—</sup>"


def test_single_molecule2():
    assert formula.molecule("CH3Cl", "html") == "CH<sub>3</sub>Cl"


def test_single_molecule3():
    assert formula.molecule("C6H12O6", "html") == "C<sub>6</sub>H<sub>1</sub><sub>2</sub>O<sub>6</sub>"


def test_reaction():
    assert formula.molecule("F- + CH3Cl", "html") == "F<sup>—</sup> + CH<sub>3</sub>Cl"


def test_reaction2():
    assert formula.molecule("F- + CH3Cl -> CH3F + Cl-", "html") == "F<sup>—</sup> + CH<sub>3</sub>Cl -> CH<sub>3</sub>F + Cl<sup>—</sup>"


if __name__ == "__main__":
    import pytest

    pytest.main()



import attrs


@attrs.define
class VDDCharge:
    atom_index: int  # index of the atom in the molecules
    atom_symbol: str  # symbol of the atom
    charge: float  # mili-electrons
    frag_index: int  # index of the fragment in the fragments list

    def change_unit(self, ratio: float):
        self.charge *= ratio



import os
from tcutility import molecule
from scm import plams

j = os.path.join


def test_atom_flags() -> None:
    xyzfile = j(os.path.split(__file__)[0], "fixtures", "xyz", "transitionstate_radical_addition.xyz")
    mol = molecule.load(xyzfile)
    atom = mol[2]
    assert atom.flags.origin == "substrate"


def test_atom_tags() -> None:
    xyzfile = j(os.path.split(__file__)[0], "fixtures", "xyz", "transitionstate_radical_addition.xyz")
    mol = molecule.load(xyzfile)
    atom = mol[3]
    assert "R1" in atom.flags.tags


def test_mol_reading() -> None:
    xyzfile = j(os.path.split(__file__)[0], "fixtures", "xyz", "transitionstate_radical_addition.xyz")
    tcmol = molecule.load(xyzfile)
    plamsmol = plams.Molecule(xyzfile)
    assert all(tcatom.symbol == plamsatom.symbol and tcatom.coords == plamsatom.coords for tcatom, plamsatom in zip(tcmol.atoms, plamsmol.atoms))


def test_mol_flags() -> None:
    xyzfile = j(os.path.split(__file__)[0], "fixtures", "xyz", "transitionstate_radical_addition.xyz")
    mol = molecule.load(xyzfile)
    assert mol.flags.task == "TransitionStateSearch"


def test_mol_tags() -> None:
    xyzfile = j(os.path.split(__file__)[0], "fixtures", "xyz", "transitionstate_radical_addition.xyz")
    mol = molecule.load(xyzfile)
    assert "vibrations" in mol.flags.tags


def test_comment() -> None:
    xyzfile = j(os.path.split(__file__)[0], "fixtures", "xyz", "transitionstate_radical_addition.xyz")
    mol = molecule.load(xyzfile)
    assert mol.comment == "molecule used for testing the molecule module"


def test_list_value() -> None:
    xyzfile = j(os.path.split(__file__)[0], "fixtures", "xyz", "pyr.xyz")
    mol = molecule.load(xyzfile)
    assert mol.flags.conn == [6, 12]


def test_list_value2() -> None:
    xyzfile = j(os.path.split(__file__)[0], "fixtures", "xyz", "pyr.xyz")
    mol = molecule.load(xyzfile)
    assert mol.flags.test_list == [3.14, "pi"]


def test_mol_copy() -> None:
    xyzfile = j(os.path.split(__file__)[0], "fixtures", "xyz", "pyr.xyz")
    mol = molecule.load(xyzfile)
    mol.copy()


def test_guess_fragments() -> None:
    xyzfile = j(os.path.split(__file__)[0], "fixtures", "xyz", "NaCl.xyz")
    mol = molecule.load(xyzfile)
    frags = molecule.guess_fragments(mol)
    assert frags['Na'][1].symbol == 'Na'


def test_guess_fragments2() -> None:
    xyzfile = j(os.path.split(__file__)[0], "fixtures", "xyz", "NaCl.xyz")
    mol = molecule.load(xyzfile)
    frags = molecule.guess_fragments(mol)
    assert frags['Na'].flags['charge'] == 1


def test_guess_fragments3() -> None:
    xyzfile = j(os.path.split(__file__)[0], "fixtures", "xyz", "NaCl_homolytic.xyz")
    mol = molecule.load(xyzfile)
    frags = molecule.guess_fragments(mol)
    assert frags['Cl'].flags['spinpol'] == 1


if __name__ == "__main__":
    import pytest

    pytest.main()



import os
from typing import Tuple

from tcutility.job.adf import ADFJob
from tcutility.job.generic import Job

j = os.path.join


class NMRJob(Job):
    """
    A job that handles calculation of Nuclear Magnetic Resonance (NMR) chemical shifts.
    The job will first run an :class:`tcutility.job.adf.ADFJob` calculation at the SAOP/TZ2P level of theory to prepare for the NMR program of AMS.
    The NMR shifts will be calculated for all atoms and any additional coordinate (NICS).
    New NICS points can be given using the :meth:`NMRJob.add_nics_point` method.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pre_nmr_job = ADFJob(*args, **kwargs)
        self.pre_nmr_job.functional("SAOP")
        self.pre_nmr_job.basis_set("TZ2P")
        self.pre_nmr_job.settings.input.adf.save = "TAPE10"
        self.nics_points = []

    def _setup_job(self):
        os.makedirs(self.workdir, exist_ok=True)

        # set up the pre_nmr_job
        self.pre_nmr_job.rundir = self.rundir
        self.pre_nmr_job.name = self.name + "_pre"
        self.pre_nmr_job.molecule(self._molecule or self._molecule_path)
        self.pre_nmr_job.sbatch(**self._sbatch)

        # we have to add the NICS points in the pre_nmr_job
        multipole_coords = []
        for point in self.nics_points:
            multipole_coords.append(f"        {point[0]} {point[1]} {point[2]} 0.0")
        self.pre_nmr_job.settings.input.ams.System.ElectrostaticEmbedding.MultipolePotential.Coordinates = "\n" + "\n".join(multipole_coords) + "\n      End"

        # we copy the pre_nmr_job output files to the main job directory
        self.pre_nmr_job.add_postamble(f'cp {j(self.pre_nmr_job.workdir, "adf.rkf")} {j(self.workdir, "TAPE21")}')
        self.pre_nmr_job.add_postamble(f'cp {j(self.pre_nmr_job.workdir, "TAPE10")} {j(self.workdir, "TAPE10")}')
        self.pre_nmr_job.run()

        # some extra settings for the main job
        self.dependency(self.pre_nmr_job)
        # NMR will name the calculation as TAPE21 instead of adf.rkf, so we rename it here
        self.add_postamble(f'mv {j(self.workdir, "TAPE21")} {j(self.workdir, "adf.rkf")}')

        # write the input for the NMR job
        # the ghost block holds the NICS points
        ghost_block = "\t"
        if len(self.nics_points) > 0:
            ghost_block += "Ghosts\n"
            for point in self.nics_points:
                ghost_block += f"    {point[0]} {point[1]} {point[2]}\n"
            ghost_block += "SubEnd\n"

        # write input and runfile
        with open(self.inputfile_path, "w+") as inpf:
            inpf.write("NMR\n")
            inpf.write("\tout all\n")
            inpf.write(ghost_block + "\n")
            inpf.write("End\n")

        with open(self.runfile_path, "w+") as runf:
            runf.write("#!/bin/sh\n\n")  # the shebang is not written by default by ADF
            runf.write("\n".join(self._preambles) + "\n\n")
            runf.write(f"$AMSBIN/nmr {self.inputfile_path} < /dev/null\n")
            runf.write("\n".join(self._postambles))

        return True

    def add_nics_point(self, coordinate: Tuple[float, float, float]):
        """
        Add a NICS point to be calculated.

        Args:
            coordinate: the (x, y, z) coordinates of a NICS point to calculate the chemical shift for. Has to be given as cartesian coordinates and using unit angstrom.
        """
        self.nics_points.append(coordinate)


if __name__ == "__main__":
    with NMRJob(test_mode=True) as job:
        job.molecule(r"D:\Users\Yuman\Desktop\PhD\TCutility\examples\job\water_dimer.xyz")
        job.add_nics_point([0, 1, 2])
        job.add_nics_point([0, 1, 2])
        job.add_nics_point([0, 1, 2])
        job.add_nics_point([0, 1, 2])
        job.add_nics_point([0, 1, 2])
        job.add_nics_point([0, 1, 2])



import os
import shutil
import stat
import subprocess as sp
from typing import List, Union

import dictfunc
from scm import plams

from tcutility import log, molecule, results, slurm
from tcutility.environment import OSName, get_os_name
from tcutility.errors import TCJobError

j = os.path.join


def _python_path():
    """
    Sometimes it is necessary to have the Python path as some environments don't have its path.
    This function attempts to find the Python path and returns it.
    """
    python = sp.run("which python", shell=True, capture_output=True).stdout.decode().strip()

    if python == "" or not os.path.exists(python):
        python = sp.run("which python3", shell=True, capture_output=True).stdout.decode().strip()

    # we default to the python executable
    if python == "" or not os.path.exists(python):
        python = "python"

    return python


class Job:
    """This is the base Job class used to build more advanced classes such as :class:`AMSJob <tcutility.job.ams.AMSJob>` and :class:`ORCAJob <tcutility.job.orca.ORCAJob>`.
    The base class contains an empty :class:`Result <tcutility.results.result.Result>` object that holds the settings.
    It also provides :meth:`__enter__` and :meth:`__exit__` methods to make use of context manager syntax.

    All class methods are in principle safe to overwrite, but the :meth:`_setup_job` method **must** be overwritten.

    Args:
        test_mode: whether to enable the testing mode. If enabled, the job will be setup like normally, but the running step is skipped. This is useful if you want to know what the job settings look like before running the real calculations.
        overwrite: whether to overwrite a previously run job in the same working directory.
        wait_for_finish: whether to wait for this job to finish running before continuing your runscript.
        delete_on_finish: whether to remove the workdir for this job after it is finished running.
    """

    def __init__(
        self, *base_jobs: List["Job"], test_mode: bool = None, overwrite: bool = None, wait_for_finish: bool = None, delete_on_finish: bool = None, delete_on_fail: bool = None, use_slurm: bool = True
    ):
        self._sbatch = results.Result()
        self._molecule = None
        self._molecule_path = None
        self.slurm_job_id = None
        self.name = "calc"
        self.rundir = "tmp"
        self._preambles = []
        self._postambles = []
        self._postscripts = []

        self.test_mode = test_mode
        self.overwrite = overwrite
        self.wait_for_finish = wait_for_finish
        self.delete_on_finish = delete_on_finish
        self.use_slurm = use_slurm

        # update this job with base_jobs
        for base_job in base_jobs:
            self.__dict__.update(base_job.copy().__dict__)

        self.test_mode = self.test_mode if test_mode is None else test_mode
        self.overwrite = self.overwrite if overwrite is None else overwrite
        self.wait_for_finish = self.wait_for_finish if wait_for_finish is None else wait_for_finish
        self.delete_on_finish = self.delete_on_finish if delete_on_finish is None else delete_on_finish
        self.delete_on_fail = delete_on_fail if delete_on_fail is None else delete_on_fail
        self.use_slurm = self.use_slurm if use_slurm is None else use_slurm

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        if exc_type:
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            log.error(f'Job set-up failed with exception: {exc_type.__name__}({exc_value}) in File "{fname}", line {exc_tb.tb_lineno}.')
            return True
        self.run()

    def can_skip(self):
        """
        Check whether the job can be skipped. We check this by loading the calculation and checking if the job status was fatal.
        Fatal in this case means that the job failed, canceled or could not be found. In those cases we want to run the job.

        .. note::
            This method works for the :class:`ADFJob <tcutility.job.adf.ADFJob>`, :class:`ADFFragmentJob <tcutility.job.adf.ADFFragmentJob>`,
            :class:`DFTBJob <tcutility.job.dftb.DFTBJob>` and :class:`ORCAJob <tcutility.job.orca.ORCAJob>` objects, but not
            yet for :class:`CRESTJob <tcutility.job.crest.CRESTJob>` and :class:`QCGJob <tcutility.job.crest.QCGJob>`. For the latter objects the job will always be rerun.
            This will be fixed in a later version of TCutility.
        """
        res = results.read(self.workdir)
        return not res.status.fatal

    def in_queue(self):
        """
        Check whether the job is currently managed by slurm.
        We check this by loading the calculation and checking if the job status is 'RUNNING', 'COMPLETING', 'CONFIGURING' or 'PENDING'.
        """
        res = results.read(self.workdir)
        return res.status.name in ["RUNNING", "COMPLETING", "CONFIGURING", "PENDING"]

    def __repr__(self):
        return f"{type(self)}(name={self.name}, rundir={self.rundir})"

    def sbatch(self, **kwargs):
        """
        Change slurm settings, for example, to change the partition or change the number of cores to use.
        The arguments are the same as you would use for sbatch (`see sbatch manual <https://slurm.schedmd.com/sbatch.html>`_). E.g. to change the partition to 'tc' call:

        ``job.sbatch(p='tc')`` or ``job.sbatch(partition='tc')``.

        Flags can be set as arguments with a boolean to enable or disable them:

        ``job.sbatch(exclusive=True)`` will set the ``--exclusive`` flag.

        .. warning::

            Note that some sbatch options, such as ``--job-name`` contain a dash, which cannot be used in Python arguments.
            To use these options you should use an underscore, like ``job.sbatch(job_name='water_dimer_GO')``.

        .. note::

            When running the job using sbatch we add a few extra default options:

            * ``-D/--chdir {self.workdir}``                to make sure the job starts in the correct directory
            * ``-J/--job-name {self.rundir}/{self.name}``  to give a nicer job-name when calling squeue
            * ``-o/--output {self.name}.out``              to redirect the output from the default slurm-{id}.out

            You can still overwrite them if you wish.
        """
        for key, value in kwargs.items():
            if key == "dependency" and "dependency" in self._sbatch:
                value = self._sbatch["dependency"] + "," + value
            self._sbatch[key] = value

    def _setup_job(self):
        """
        Set up the current job. This method should create the working directory, runscript and input file.
        Method must return True if it was successful.
        """
        raise NotImplementedError("You must implement the _setup_job method in your subclass.")

    def run(self):
        """
        Run this job. We detect if we are using slurm. If we are we submit this job using sbatch. Otherwise, we will run the job locally.
        """
        if self.overwrite:
            shutil.rmtree(self.workdir) if os.path.exists(self.workdir) else None
            os.makedirs(self.workdir, exist_ok=True)

        if self.can_skip():
            log.info(f"Skipping calculation {j(self.rundir, self.name)}, it is already finished or currently pending or running.")
            return

        # write the post-script calls to the post-ambles:
        if self.delete_on_finish:
            self.add_postamble(f"rm -r {self.workdir}")

        if self.delete_on_fail:
            self.add_postamble("# this will delete the calculation if it failed")
            self.add_postamble(f"if [[ `tc read -s {self.workdir}` = FAILED || `tc read -s {self.workdir}` = UNKNOWN ]]; then rm -r {self.workdir}; fi;")

        for postscript in self._postscripts:
            self._postambles.append(f'{_python_path()} {postscript[0]} {" ".join(postscript[1])}')

        # setup the job and check if it was successfull
        setup_success = self._setup_job()

        if self.test_mode or not setup_success:
            return

        if slurm.has_slurm() and self.use_slurm:
            # set some default sbatch settings
            if any(option not in self._sbatch for option in ["D", "chdir"]):
                self._sbatch.setdefault("D", self.workdir)
            if any(option not in self._sbatch for option in ["J", "job_name"]):
                self._sbatch.setdefault("J", f"{self.rundir}/{self.name}")
            if any(option not in self._sbatch for option in ["o", "output"]):
                self._sbatch.setdefault("o", f"{self.name}.out")
            self._sbatch.prune()

            # submit the job with sbatch
            sbatch_result = slurm.sbatch(os.path.split(self.runfile_path)[1], **self._sbatch)

            # store the slurm job ID
            self.slurm_job_id = sbatch_result.id
            # and write the command to a file so we can rerun it later
            with open(j(self.workdir, "submit.sh"), "w+") as cmd_file:
                cmd_file.write(sbatch_result.command)
            # make the submit command executable
            os.chmod(j(self.workdir, "submit.sh"), stat.S_IRWXU)

            # if we requested the job to hold we will wait for the slurm job to finish
            if self.wait_for_finish:
                slurm.wait_for_job(self.slurm_job_id)
        else:
            os_name = get_os_name()

            if os_name == OSName.WINDOWS:
                raise TCJobError("Generic Job", "Running jobs on Windows is not supported.")

            # if we are not using slurm, we can execute the file. For this we need special permissions, so we have to set that first.
            os.chmod(self.runfile_path, os.stat(self.runfile_path).st_mode | stat.S_IEXEC)

            runfile_dir, runscript = os.path.split(self.runfile_path)
            command = ["./" + runscript] if os.name == "posix" else ["sh", runscript]
            print(f"Running command: {command} in directory: {runfile_dir}")

            with open(f"{os.path.split(self.runfile_path)[0]}/{self.name}.out", "w+") as out:
                sp.run(command, cwd=runfile_dir, stdout=out, shell=True)

    def add_preamble(self, line: str):
        """
        Add a preamble for the runscript. This should come after the shebang, but before the calculation is ran by the program (ADF or ORCA).
        This can used, for example, to load some modules. E.g. to load a specific version of AMS we can call:
        job.add_preamble('module load ams/2023.101')
        """
        self._preambles.append(line)

    def add_postamble(self, line: str):
        """
        Add a postamble for the runscript. This should come after the calculation is ran by the program (ADF or ORCA).
        This can be used, for example, to remove or copy some files. E.g. to remove all t12.* files we can call:
        job.add_postamble('rm t12.*')
        """
        self._postambles.append(line)

    def add_postscript(self, script, *args):
        """
        Add a post-script to this calculation.
        This should be either a Python module with a __file__ attribute or the path to a Python script.
        The post-script will be called with Python and any given args will be added as arguments when calling the script.

        Args:
            script: a Python object with a __file__ attribute or the file-path to a script.
            *args: positional arguments to pass to the post-script.
        """
        if not isinstance(script, str):
            script = script.__file__
        self._postscripts.append((script, args))

    def dependency(self, otherjob: "Job"):
        """
        Set a dependency between this job and otherjob.
        This means that this job will run after the other job is finished running succesfully.
        """
        if otherjob.can_skip() and not otherjob.in_queue():
            return

        if hasattr(otherjob, "slurm_job_id"):
            self.sbatch(dependency=f"afterok:{otherjob.slurm_job_id}")
            self.sbatch(kill_on_invalid_dep="Yes")

    @property
    def workdir(self):
        """
        The working directory of this job. All important files are written here, for example the input file and runscript.
        """
        return j(os.path.abspath(self.rundir), self.name)

    @property
    def runfile_path(self):
        """
        The file path to the runscript of this job.
        """
        return j(self.workdir, f"{self.name}.run")

    @property
    def inputfile_path(self):
        """
        The file path to the input file of this job.
        """
        return j(self.workdir, f"{self.name}.in")

    @property
    def output_mol_path(self):
        """
        This method should return the name of the output molecule if it makes sense to give it back.
        E.g. for ADF it will be output.xyz in the workdir for optimization jobs.
        """
        raise NotImplementedError("You must implement the _setup_job method in your subclass.")

    def molecule(self, mol: Union[str, plams.Molecule, plams.Atom, List[plams.Atom]]):
        """
        Add a molecule to this calculation in various formats.

        Args:
            mol: the molecule to read, can be a path (str). If the path exists already we read it. If it does not exist yet, it will be read in later. mol can also be a plams.Molecule object or a single or a list of plams.Atom objects.
        """
        if isinstance(mol, plams.Molecule):
            self._molecule = mol

        elif isinstance(mol, str) and os.path.exists(mol):
            self._molecule = molecule.load(mol)

        elif isinstance(mol, str):
            self._molecule_path = os.path.abspath(mol)

        elif isinstance(mol, list) and isinstance(mol[0], plams.Atom):
            self._molecule = plams.Molecule()
            [self._molecule.add_atom(atom) for atom in mol]

        elif isinstance(mol, plams.Atom):
            self._molecule = plams.Molecule()
            self._molecule.add_atom(mol)

    def copy(self):
        """
        Make and return a copy of this object.
        """
        import copy

        cp = Job()
        # cast this object to a list of keys and values
        lsts = dictfunc.dict_to_list(self.__dict__)
        # copy everthing in the lists
        lsts = [[copy.copy(x) for x in lst] for lst in lsts]
        # and return a new result object
        cp.__dict__.update(results.Result(dictfunc.list_to_dict(lsts)))
        return cp



from tcutility import results, constants
import os

j = os.path.join


def test_read_orca() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "orca", "optimization"))
    assert isinstance(res, results.Result)


def test_read_orca2() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "orca", "sp_freq"))
    assert isinstance(res, results.Result)


def test_read_engine() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "orca", "optimization"))
    assert res.engine == "orca"


def test_read_engine2() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "orca", "sp_freq"))
    assert res.engine == "orca"


def test_LOT() -> None:
    # LOT = level of theory
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "orca", "optimization"))
    assert res.level.summary == "QRO-CCSD(T)/cc-pVTZ"


def test_character() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "orca", "sp_freq"))
    assert res.properties.vibrations.character == "transitionstate"


def test_imag_freq() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "orca", "sp_freq"))
    assert res.properties.vibrations.frequencies[0] == -47.03


def test_hf_energy() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "orca", "optimization"))
    assert round(res.properties.energy.HF, 2) == round(-132.433551362 * constants.HA2KCALMOL, 2)


def test_mp2_energy() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "orca", "optimization"))
    assert round(res.properties.energy.MP2, 2) == round(-132.935938324 * constants.HA2KCALMOL, 2)


def test_ccsd_energy() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "orca", "optimization"))
    assert round(res.properties.energy.CCSD, 2) == round(-132.964383423 * constants.HA2KCALMOL, 2)


def test_ccsd_t_energy() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "orca", "optimization"))
    assert round(res.properties.energy.CCSD_T, 2) == round(-132.986982793 * constants.HA2KCALMOL, 2)


def test_t1() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "orca", "optimization"))
    assert round(res.properties.t1, 5) == round(0.013815962, 5)


def test_s2() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "orca", "optimization"))
    assert round(res.properties.s2, 5) == round(0.759369, 5)


def test_spin_contam() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "orca", "optimization"))
    assert round(res.properties.spin_contamination * 100, 2) == 1.25


if __name__ == "__main__":
    import pytest

    pytest.main()



from time import perf_counter
import numpy as np
from tcutility import log
from types import FunctionType
import listfunc
import atexit
'''
Implements a decorator and context manager that records and stores the number of times a function
has been called and also the amount of time taken in total/per call
'''
times = {}
exec_start = perf_counter()

enabled: bool = True


class timer:
    '''
    The main timer class. It acts both as a context-manager and decorator.
    '''
    def __init__(self, name=None):
        if isinstance(name, FunctionType):
            self.function = name
            times.setdefault(name.__qualname__, {'calls': 0, 'timings': []})
        else:
            self.name = name
            times.setdefault(self.name, {'calls': 0, 'timings': []})

    def __enter__(self):
        self.start = perf_counter()
        return self

    def __exit__(self, *args, **kwargs):
        times[self.name]['calls'] += 1
        times[self.name]['timings'].append(perf_counter() - self.start)

    def __call__(self, *args, **kwargs):
        times.setdefault(self.function.__qualname__, {'calls': 0, 'timings': []})

        if enabled and __debug__:
            start = perf_counter()
            ret = self.function(*args, **kwargs)
            times[self.function.__qualname__]['calls'] += 1
            times[self.function.__qualname__]['timings'].append(perf_counter() - start)
            return ret
        else:
            return self.function(*args, **kwargs)


def print_timings():
    if not enabled:
        return

    if sum(time['calls'] for time in times.values()) == 0:
        return

    names = list(times.keys())
    names = sorted(names)
    names_splits = [name.split('.') for name in names]

    # generate names that are common to multiple names
    for i, name in enumerate(names_splits):
        parents = []
        for name_ in names_splits[:i][::-1]:
            parents.append(".".join(listfunc.common_list(name, name_)))

        parents = [parent for parent in parents if parent]
        if parents != []:
            longest_parent = sorted(parents, key=lambda x: len(x))[-1]
            if longest_parent not in names:
                names.append(longest_parent)

    names = sorted(names)
    names_splits = [name.split('.') for name in names]
    # get the parents:
    parents = []
    parent_levels = []
    for i, (name, splits) in enumerate(zip(names, names_splits)):
        parent = 'TOTAL'
        parent_level = 0
        for name_, splits_ in zip(names[:i], names_splits[:i]):
            if listfunc.startswith(splits, splits_):
                parent = name_
                parent_level += 1
        parents.append(parent)
        parent_levels.append(parent_level)

    parent_times = {parent: {'timings': [sum([sum(times.get(name, {'timings': [0]})['timings']) for name in names if name.startswith(parent)])]} for parent in parents}
    for parent in parents:
        if parent in times:
            parent_times[parent] = {'timings': times[parent]['timings'], 'calls': times[parent]['calls']}
    parent_times['TOTAL'] = {'timings': [perf_counter() - exec_start], 'calls': 1}

    header = ['Function', 'Calls', 'Mean (s)', 'Time Spent (s)', 'Rel. Time']
    times_ = parent_times.copy()
    times_.update(times)
    lines = []
    for name, parent, level in zip(names, parents, parent_levels):
        line = []
        line.append(name.replace(parent, ' > ' * level))  # function name
        calls = times_[name].get("calls", 0)
        if calls == 0:
            line.append('')
            line.append('')
            line.append('')
        else:
            mean = np.mean(times_[name]["timings"])
            total = np.sum(times_[name]["timings"])
            line.append(str(calls))
            line.append(f'{mean:.3f}')
            line.append(f'{total:.3f}')

        try:
            rel = sum(times_[name]["timings"])/sum(parent_times[parent]["timings"])*100
            rel_total = sum(times_[name]["timings"])/sum(parent_times["TOTAL"]["timings"])*100
        except ZeroDivisionError:
            rel = 0
            rel_total = 0
        if parent != 'TOTAL':
            line.append(f'{rel_total: >3.0f}% ({rel: >3.0f}%)')
        else:
            line.append(f'{rel_total: >3.0f}%')

        lines.append(line)
    lines.append(['TOTAL', '', '', f'{sum(times_["TOTAL"]["timings"]):.3f}', '100%'])
    
    log.table(lines, header=header, hline=[-2])


# this makes sure that the timings are printed when Python quits running
atexit.register(print_timings)



from tcutility.job.ams import AMSJob
from tcutility import results
import os


j = os.path.join


class DFTBJob(AMSJob):
    '''
    Setup and run a density functional with tight-binding (DFTB) calculation as implemented in the Amsterdam modelling suite (AMS).
    This class supports all methods of the parent :class:`AMSJob <tcutility.job.ams.AMSJob>`.
    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.settings = results.Result()
        self._model = None
        self.kspace('Good')
        self.single_point()
        self.model('GFN1-xTB')
        self.solvent('vacuum')

    def kspace(self, quality: str = 'Good'):
        '''
        Set the k-space integration quality for this job.

        Args:
            quality: the type of basis-set to use. Default is ``Good``.
        '''
        self.settings.input.DFTB.kspace.quality = quality

    def model(self, name: str = 'GFN1-xTB', dispersion: str = None, parameter_dir: str = None):
        '''
        Set the model Hamiltonian for the job to use.

        Args:
            name: name of the model Hamiltonian. This is the same name as the one in the DFTB gui. Default is ``GFN1-xTB``.
        '''
        self.settings.input.DFTB.Model = name

        if dispersion is None:
            for disp_suffix in ['-UFF', '-ULG', '-D2', '-D3-BJ', '-D4', '-Auto']:
                if not name.endswith(disp_suffix):
                    continue
                self.settings.input.DFTB.DispersionCorrection = disp_suffix[1:]
                self.settings.input.DFTB.Model[:-len(disp_suffix)]
        else:
            self.settings.input.DFTB.DispersionCorrection = dispersion

        if parameter_dir:
            self.settings.input.DFTB.ResourcesDir = parameter_dir

    def solvent(self, name: str = None, grid_size=974):
        '''
        Model solvation using the GBSA model.

        Args:
            name: the name of the solvent you want to use. Must be ``None``, ``Acetone``, ``Acetonitrile``, ``CHCl3``, ``CS2``, ``DMSO``, ``Ether``, ``H2O``, ``Methanol``, ``THF`` or ``Toluene``.
            grid_size: the size of the grid used to construct the solvent accessible surface. Must be ``230``, ``974``, ``2030`` or ``5810``.
        '''
        if name == 'vacuum':
            self.settings.input.DFTB.pop('solvation', None)
            return

        self.settings.input.DFTB.Solvation.Solvent = name
        self.settings.input.DFTB.Solvation.SurfaceGrid = grid_size


if __name__ == '__main__':
    with DFTBJob(test_mode=False, overwrite=True) as job:
        job.rundir = 'tmp/SN2'
        job.name = 'DFTB'
        job.molecule('../../../test/fixtures/xyz/transitionstate_radical_addition.xyz')
        job.sbatch(p='tc', ntasks_per_node=15)
        job.model('GFN1-xTB')
        job.optimization()
        job.kspace('Good')
        job.solvent('H2O')



"""Module containing functions for generating citations"""

import os
from math import ceil
from typing import List

import click
import docx
import htmldocx
from docx.enum.text import WD_ALIGN_PARAGRAPH
from docx.shared import Pt
from tcutility import cite, data, spell_check


class Docx:
    """
    Small class that handles writing to a docx file. This should and will be moved to its own module in TCutility soon.
    """

    def __init__(self, file="test.docx", overwrite=False):
        self.file = file
        if not os.path.exists(file) or overwrite:
            self.doc = docx.Document()
        else:
            self.doc = docx.Document(file)

        self.doc.styles["Normal"].font.name = "Times New Roman"
        self.doc.styles["Normal"].font.size = Pt(12)
        self.doc.styles["Normal"].paragraph_format.space_after = 0
        self.html_parser = htmldocx.HtmlToDocx()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.doc.save(self.file)

    def write_paragraph(self, text, alignment=WD_ALIGN_PARAGRAPH.LEFT):
        """
        Write a piece of text as a pragraph to this Docx file.
        This function will parse any HTML that is given in the text.
        E.g. you can use the <b></b> tags to make a piece of text bold.
        """
        self.html_parser.add_html_to_document(text, self.doc)
        self.doc.paragraphs[-1].alignment = alignment

    def open(self):
        """
        Open this file in Word.
        """
        os.system(f"open {self.file}")


program_references = {
    "ams": [
        "10.1002/jcc.1056",
        "10.1007/s002140050353",
    ],
    "adf": [
        "10.1002/jcc.1056",
        "10.1007/s002140050353",
    ],
    "orca": [
        "10.1002/wcms.81",
        "10.1063/5.0004608",
        "10.1002/wcms.1606",
    ],
    "dftb": [],
    "xtb": [],
    "cosmo": [],
    "crest": [],
    "pyfrag": [
        "10.1002/jcc.20786",
        "10.1002/jcc.25871",
        "10.5281/zenodo.1045523",
    ],
    "pyorb": [],
    "cylview": [],
}

methodology_references = {
    "fmatsfo": [],
    "eda": ["10.1002/9780470125922.ch1", "10.1515/9783110660074-008"],
    "asm": [
        "10.1038/s41596-019-0265-0",
        "10.1002/(SICI)1096-987X(19990115)20:1<114::AID-JCC12>3.0.CO;2-L",
        "10.1039/B926828F",
        "10.1039/C4CS00055B",
        "10.1002/wcms.1221",
        "10.1002/anie.201701486",
        "10.1039/D1CC02042K",
    ],
    "zora": [
        "10.1063/1.466059",
        "10.1063/1.467943",
    ],
    "ksmo": [],
    "vdd": [
        "10.1002/jcc.10351",
        "10.1002/jcc.27184",
        "10.1039/c5cp07483e",
    ],
    "hydrogen-bonding": [],
    "halogen-bonding": [],
    "chalcogen-bonding": [],
    "pnictogen-bondding": [],
    "zlm-fit": [
        "10.1021/ct500172n",
    ],
    "becke-grid": [
        "1.1002/jcc.23323",
    ],
    "sto": [
        "10.1002/jcc.10255",
    ],
    "vibrational-analysis": [
        "10.1016/s0010-4655(96)00120-8",
        "10.1016/S0010-4655(96)00119-1",
        "10.1002/qua.20653",
    ],
    "irc": [],
    "d3": ["10.1063/1.3382344"],
    "d3bj": ["10.1002/jcc.21759"],
    "d4": ["10.1063/1.5090222"],
}


doi_order = []


def format_paragraph(dois, style):
    paragraphs = []
    for doi in dois:
        if doi not in doi_order:
            doi_order.append(doi)
        try:
            citation = cite.cite(doi, style=style, mode="plain")
        except Exception as exp:
            print("\t" + str(exp))
            raise exp

        print(f"  [{doi}] {citation}")
        paragraph = f'[{doi_order.index(doi) + 1}] {cite.cite(doi, style=style, mode="html")}'
        paragraphs.append(paragraph)

    return paragraphs


def _print_rect_list(printables, spaces_before=0):
    """
    This function prints a list of strings in a rectangle to the output.
    This is similar to what the ls program does in unix.
    """
    n_shell_col = os.get_terminal_size().columns
    prev_mat = []
    prev_col_lens = []

    # we first have to determine the correct dimensions of our rectangle
    for ncol in range(1, n_shell_col):
        # the number of rows for the number of columns
        nrows = ceil(len(printables) / ncol)
        # we then get what the rectangle would be
        mat = [printables[i * ncol : (i + 1) * ncol] for i in range(nrows)]
        # and determine for each column the width
        col_lens = [max([len(row[i]) for row in mat if i < len(row)] + [0]) for i in range(ncol)]
        # then calculate the length of each row based on the column lengths
        # we use a spacing of 2 spaces between each column
        row_len = spaces_before + sum(col_lens) + 2 * len(col_lens) - 2

        # if the rows are too big we exit the loop
        if row_len > n_shell_col:
            break

        # store the previous loops results
        prev_col_lens = col_lens
        prev_mat = mat

    # then print the strings with the right column widths
    for row in prev_mat:
        print(" " * spaces_before + "  ".join([x.ljust(col_len) for x, col_len in zip(row, prev_col_lens)]))


@click.command()
@click.argument("objects", type=str, nargs=-1, required=False)
@click.option("-w", "--wiley", help="Set the citation style to Wiley. This is the default style.", flag_value="wiley", default=True)
@click.option("-a", "--acs", help="Set the citation style to ACS.", flag_value="acs")
@click.option("-r", "--rsc", help="Set the citation style to RSC.", flag_value="rsc")
@click.option("-o", "--output", help="The output Word file to write the citations to.", type=str, default="citations.docx")
@click.option("-l", "--list_citations", help="List currently available citations.", is_flag=True, default=False)
def generate_citations(objects: List[str], wiley: bool, acs: bool, rsc: bool, output: str, list_citations: bool) -> None:
    """
    Generate citations for objects.

    Currently supports generating citations for functionals, basis-sets, programs, methodologies and DOIs.
    This program also generates, and if possible, opens a Word document that contains the formatted citations.
    Multiple objects can be given separated by spaces.
    If the supplied object is also a file path it will read each line as a separate object.

    Example usage:

    > tc cite ADF
    Program ORCA
      [10.1002/wcms.81] F. Neese, WIREs Comput. Mol. Sci. 2011, 2, 73-78.
      [10.1063/5.0004608] F. Neese, F. Wennmohs, U. Becker, C. Riplinger, J. Chem. Phys. 2020, 152.
      [10.1002/wcms.1606] F. Neese, WIREs Comput. Mol. Sci. 2022, 12.

    > tc cite BP86 BLYP OLYP OPBE D3BJ
    Functional BP86
      [10.1103/PhysRevA.38.3098] A. D. Becke, Phys. Rev. A 1988, 38, 3098-3100.
      [10.1103/PhysRevB.33.8800] J. P. Perdew, W. Yue, Phys. Rev. B 1986, 33, 8800-8802.
    Functional BLYP
      [10.1103/PhysRevA.38.3098] A. D. Becke, Phys. Rev. A 1988, 38, 3098-3100.
      [10.1103/PhysRevB.37.785] C. Lee, W. Yang, R. G. Parr, Phys. Rev. B 1988, 37, 785-789.
    Functional OLYP
      [10.1080/00268970010018431] N. C. Handy, A. J. Cohen, Mol. Phys. 2001, 99, 403-412.
      [10.1103/PhysRevB.37.785] C. Lee, W. Yang, R. G. Parr, Phys. Rev. B 1988, 37, 785-789.
    Functional OPBE
      [10.1080/00268970010018431] N. C. Handy, A. J. Cohen, Mol. Phys. 2001, 99, 403-412.
      [10.1103/PhysRevLett.77.3865] J. P. Perdew, K. Burke, M. Ernzerhof, Phys. Rev. Lett. 1996, 77, 3865-3868.
    Methodology D3BJ
      [10.1002/jcc.21759] S. Grimme, S. Ehrlich, L. Goerigk, J. Comput. Chem. 2011, 32, 1456-1465.
    """
    available_citations = list(program_references.keys()) + list(methodology_references.keys()) + list(data.functionals.functionals.keys())
    if list_citations:
        print("OBJECTS WITH AVAILABLE REFERENCES:")
        print("    Programs")
        print("    ========")
        printables = [prog for prog, dois in program_references.items() if len(dois) > 0]
        _print_rect_list(printables, 4)
        print()

        print("    Methodology")
        print("    ===========")
        printables = [meth for meth, dois in methodology_references.items() if len(dois) > 0]
        _print_rect_list(printables, 4)
        print()

        print("    Functionals")
        print("    ===========")
        printables = [xc_info.path_safe_name for xc, xc_info in data.functionals.functionals.items() if len(xc_info.dois) > 0]  # type: ignore
        _print_rect_list(printables, 4)
        print()

        return

    if len(objects) == 1 and os.path.isfile(objects[0]):
        with open(objects[0]) as inp:
            objects = [line.strip() for line in inp.readlines()]

    style = "wiley" if wiley else "acs" if acs else "rsc"

    with Docx(file=output, overwrite=True) as out:
        for obj in objects:
            paragraphs = None  # try to format a functional
            paragraph_title = ""
            try:
                xc_info = data.functionals.get(obj)
                print("Functional", obj)
                paragraph_title = f"XC-Functional: <b>{xc_info.name_html}</b>"
                paragraphs = format_paragraph(xc_info.dois, style=style)

            except KeyError:
                pass

            # if its not a functional we look in the programs
            if obj.lower() in program_references:
                print("Program", obj)
                paragraph_title = f"Program: <b>{obj}</b>"
                paragraphs = format_paragraph(program_references[obj.lower()], style=style)

            # and the methodologies
            if obj.lower() in methodology_references:
                print("Methodology", obj)
                paragraph_title = f"Method: <b>{obj}</b>"
                paragraphs = format_paragraph(methodology_references[obj.lower()], style=style)

            # if we still dont have a paragraphs we check if it is a DOI
            if paragraphs is None and obj.startswith("10."):
                print("DOI")
                paragraph_title = f"DOI: <b>{obj}</b>"
                paragraphs = format_paragraph([obj], style=style)

            if paragraphs is None:
                spell_check.make_suggestion(obj, available_citations, ignore_case=True)
                continue

            out.write_paragraph(paragraph_title)
            for paragraph in paragraphs:
                out.write_paragraph(paragraph, alignment=WD_ALIGN_PARAGRAPH.JUSTIFY)
            out.write_paragraph(" ")

    out.open()



import pathlib as pl


# read data
data_dir = pl.Path(__file__).parents[0] / "_atom_data_info"

with open(data_dir / "name.txt") as data:
    lines = data.readlines()
_element_order = [line.split(",")[1].strip() for line in lines]

with open(data_dir / "symbol.txt") as data:
    lines = data.readlines()
_symbol_order = [line.split(",")[1].strip() for line in lines]

with open(data_dir / "radius.txt") as data:
    lines = data.readlines()
_radii = {int(line.split(",")[0]): float(line.split(",")[1].strip()) for line in lines}

with open(data_dir / "color.txt") as data:
    lines = data.readlines()
_colors = {int(line.split(",")[0]): [int(x.strip()) for x in line.split(",")[1:]] for line in lines}


def parse_element(val):
    """
    Parse a str or int to an atom number.

    Args:
        val: Element name, symbol or atom number.

    Returns:
        Atom number corresponding to val.

    Examples:

        .. code-block:: python

            parse_element('Hydrogen') == 1
            parse_element('C') == 6
            parse_element(23) == 23
    """
    # we will assume an integer value is an atom number already
    if isinstance(val, int):
        return val
    # if it is not an int it should be a string
    if val.lower() in _element_order:
        # first try to get it in the element name list
        return _element_order.index(val.lower())
    if val in _symbol_order:
        # alternatively try to get it in the symbol list
        return _symbol_order.index(val)
    raise KeyError(f'Element "{val}" not parsable.')


def radius(element):
    '''
    Args:
        element: the symbol, name or atom number of the element. See :func:`parse_element`.
    Return: 
        The empirical covalent radius of an element in angstroms, up to element 96.
    '''
    num = parse_element(element)
    return _radii.get(num)


def color(element, mode: str = 'rgb'):
    '''
    Args:
        element: the symbol, name or atom number of the element. See :func:`parse_element`.
        mode: the type of color to return. Can be `rgb` or `hex`.

    Return:
        The standard CPK colors of the elements, up to element 109. 
        If the mode is `rgb` returns a tuple of RGB values between `0` and `255`.
    '''
    num = parse_element(element)
    c = _colors[num]
    if mode == 'rgb':
        return c
    if mode == 'hex':
        return '#%02x%02x%02x' % tuple(c)


def atom_number(element):
    return parse_element(element)


def symbol(element):
    num = parse_element(element)
    return _symbol_order[num]


def element(element):
    num = parse_element(element)
    return _element_order[num]


if __name__ == "__main__":
    print(symbol(0))



from tcutility import geometry
import numpy as np
from scm import plams


def test_rotation_matrix():
    test_rot = np.array([[-0.3587314, -0.0511359, 0.9320391], [-0.2293407, -0.9630634, -0.1411088], [0.9048285, -0.2643747, 0.3337536]]).round(5)
    assert (geometry.get_rotmat(0.4, 1.2, 3).round(5) == test_rot).all()


def test_apply_rotmat():
    R = geometry.get_rotmat(y=np.pi)
    a = np.array([1, 0, 0])
    assert (geometry.apply_rotmat(a, R).round(5) == np.array([-1, 0, 0])).all()


def test_apply_rotmat2():
    R = geometry.get_rotmat()
    a = np.array([1, 0, 0])
    assert (geometry.apply_rotmat(a, R).round(5) == a).all()


def test_apply_rotmat3():
    R = geometry.get_rotmat(y=np.pi)
    a = np.array([[1, 0, 0], [1, 0, 0], [1, 0, 0], [1, 0, 0]])
    b = np.array([[-1, 0, 0], [-1, 0, 0], [-1, 0, 0], [-1, 0, 0]])
    assert (geometry.apply_rotmat(a, R).round(5) == b).all()


def test_apply_rotate():
    a = np.array([1, 0, 0])
    assert (geometry.rotate(a, y=np.pi).round(5) == np.array([-1, 0, 0])).all()


def test_apply_rotate2():
    a = np.array([1, 0, 0])
    assert (geometry.rotate(a).round(5) == a).all()


def test_apply_rotate3():
    a = np.array([[1, 0, 0], [1, 0, 0], [1, 0, 0], [1, 0, 0]])
    b = np.array([[-1, 0, 0], [-1, 0, 0], [-1, 0, 0], [-1, 0, 0]])
    assert (geometry.rotate(a, y=np.pi).round(5) == b).all()


def test_vector_align_rotmat():
    a = np.random.rand(3) * 20 - 10
    b = np.random.rand(3) * 20 - 10

    R = geometry.vector_align_rotmat(a, b)

    a = a / np.linalg.norm(a)
    b = b / np.linalg.norm(b)

    assert (geometry.apply_rotmat(a, R).round(5) == b.round(5)).all()


def test_vector_align_rotmat2():
    a = np.array([1, 0, 0])
    b = np.array([1, 0, 0])

    R = geometry.vector_align_rotmat(a, b)
    assert (R.round(5) == np.eye(3)).all()


def test_vector_align_rotmat3():
    a = np.array([-1, 0, 0])
    b = np.array([1, 0, 0])

    R = geometry.vector_align_rotmat(a, b)
    assert (geometry.apply_rotmat(a, R).round(5) == b.round(5)).all()


def test_transform():
    # create two arrays that are the same
    X, Y = np.arange(5 * 3).reshape(5, 3), np.arange(5 * 3).reshape(5, 3)  

    # create a transformation matrix to change X
    Tx = geometry.Transform()
    Tx.rotate(x=1, y=1, z=1)
    Tx.translate(x=1, y=1, z=1)

    X = Tx(X)
    
    # get the Kabsch transformation matrix
    Tkabsch = geometry.KabschTransform(X, Y)

    # check if applying the transformation matrix to X yields Y
    assert np.isclose(Tkabsch(X), Y).all()


def test_transform2():
    # create two arrays that are the same
    X, Y = np.arange(5 * 3).reshape(5, 3), np.arange(5 * 3).reshape(5, 3)  

    # get the Kabsch transformation matrix
    Tkabsch = geometry.KabschTransform(X, Y)

    # check if applying the transformation matrix to X yields Y
    assert np.isclose(Tkabsch(X), Y).all()


def test_transform_mol():
    inp = """H       0.00000000       0.00000000       0.38278869
             H       0.00000000       0.00000000      -0.38278869"""

    mol = plams.Molecule()
    for line in inp.splitlines():
        symbol, x, y, z = line.strip().split()
        mol.add_atom(plams.Atom(symbol=symbol, coords=[x, y, z]))

    # translate the molecule
    T = geometry.Transform()
    T.translate([0, 0, 0.38278869])

    # check if the second atom is centered on the origin
    assert T(mol).atoms[1].coords == (0.0, 0.0, 0.0)


def test_transform_mol2():
    inp = """H       0.00000000       0.00000000       0.38278869
             H       0.00000000       0.00000000      -0.38278869"""

    mol = plams.Molecule()
    for line in inp.splitlines():
        symbol, x, y, z = line.strip().split()
        mol.add_atom(plams.Atom(symbol=symbol, coords=[x, y, z]))

    # reflect the molecule on the yz-plane
    T = geometry.Transform()
    T.reflect([0, 0, 1])
    # check if the second atom is not in the first atoms position
    assert T(mol).atoms[1].coords == (0.0, 0.0, 0.38278869)



if __name__ == "__main__":
    import pytest

    pytest.main()
